from pyinfra import logger
from pyinfra.operations import *  # noqa: F401,F403
from pyinfra.operations import __all__

__all__ = __all__

logger.warning('Use of `pyinfra.modules` is deprecated, please use `pyinfra.operations`.')



from __future__ import unicode_literals

import re

import six

from pyinfra.api import FactBase

from .util.packaging import parse_packages


class DebPackages(FactBase):
    '''
    Returns a dict of installed dpkg packages:

    .. code:: python

        {
            'package_name': ['version'],
        }
    '''

    command = 'dpkg -l'
    requires_command = 'dpkg'

    default = dict

    regex = r'^ii\s+([a-zA-Z0-9\+\-\.]+):?[a-zA-Z0-9]*\s+([a-zA-Z0-9:~\.\-\+]+).+$'

    def process(self, output):
        return parse_packages(self.regex, output)


class DebPackage(FactBase):
    '''
    Returns information on a .deb archive or installed package.
    '''

    _regexes = {
        'name': r'^Package: ([a-zA-Z0-9\-]+)$',
        'version': r'^Version: ([0-9\:\.\-]+)$',
    }

    requires_command = 'dpkg'

    def command(self, name):
        return '! test -e {0} || (dpkg -I {0} 2> /dev/null || dpkg -s {0})'.format(name)

    def process(self, output):
        data = {}

        for line in output:
            line = line.strip()
            for key, regex in six.iteritems(self._regexes):
                matches = re.match(regex, line)
                if matches:
                    value = matches.group(1)
                    data[key] = value
                    break

        return data



'''
Manage git repositories and configuration.
'''

from __future__ import unicode_literals

import re

from pyinfra import logger
from pyinfra.api import operation, OperationError

from . import files, ssh
from .util.files import chown


@operation(pipeline_facts={
    'git_config': 'repo',
})
def config(
    key, value,
    repo=None,
    state=None, host=None,
):
    '''
    Manage git config for a repository or globally.

    + key: the key of the config to ensure
    + value: the value this key should have
    + repo: specify the git repo path to edit local config (defaults to global)

    Example:

    .. code:: python

        git.config(
            name='Ensure user name is set for a repo',
            key='user.name',
            value='Anon E. Mouse',
            repo='/usr/local/src/pyinfra',
        )

    '''

    existing_config = host.fact.git_config(repo)

    if not existing_config or existing_config.get(key) != value:
        if repo is None:
            yield 'git config --global {0} "{1}"'.format(key, value)
        else:
            yield 'cd {0} && git config --local {1} "{2}"'.format(repo, key, value)

        existing_config[key] = value

    else:
        host.noop('git config {0} is set to {1}'.format(key, value))


@operation(pipeline_facts={
    'git_branch': 'target',
})
def repo(
    src, dest,
    branch='master', pull=True, rebase=False,
    user=None, group=None, ssh_keyscan=False,
    update_submodules=False, recursive_submodules=False,
    state=None, host=None,
):
    '''
    Clone/pull git repositories.

    + src: the git source URL
    + dest: directory to clone to
    + branch: branch to pull/checkout
    + pull: pull any changes for the branch
    + rebase: when pulling, use ``--rebase``
    + user: chown files to this user after
    + group: chown files to this group after
    + ssh_keyscan: keyscan the remote host if not in known_hosts before clone/pull
    + update_submodules: update any git submodules
    + recursive_submodules: update git submodules recursively

    Example:

    .. code:: python

        git.repo(
            name='Clone repo',
            src='https://github.com/Fizzadar/pyinfra.git',
            dest='/usr/local/src/pyinfra',
        )
    '''

    # Ensure our target directory exists
    yield files.directory(dest, state=state, host=host)

    # Do we need to scan for the remote host key?
    if ssh_keyscan:
        # Attempt to parse the domain from the git repository
        domain = re.match(r'^[a-zA-Z0-9]+@([0-9a-zA-Z\.\-]+)', src)

        if domain:
            yield ssh.keyscan(domain.group(1), state=state, host=host)
        else:
            raise OperationError(
                'Could not parse domain (to SSH keyscan) from: {0}'.format(src),
            )

    # Store git commands for directory prefix
    git_commands = []
    is_repo = host.fact.directory('/'.join((dest, '.git')))

    # Cloning new repo?
    if not is_repo:
        git_commands.append('clone {0} --branch {1} .'.format(src, branch))

    # Ensuring existing repo
    else:
        current_branch = host.fact.git_branch(dest)
        if current_branch != branch:
            git_commands.append('fetch')  # fetch to ensure we have the branch locally
            git_commands.append('checkout {0}'.format(branch))

        if pull:
            if rebase:
                git_commands.append('pull --rebase')
            else:
                git_commands.append('pull')

    if update_submodules:
        if recursive_submodules:
            git_commands.append('submodule update --init --recursive')
        else:
            git_commands.append('submodule update --init')

    # Attach prefixes for directory
    command_prefix = 'cd {0} && git'.format(dest)
    git_commands = [
        '{0} {1}'.format(command_prefix, command)
        for command in git_commands
    ]

    for cmd in git_commands:
        yield cmd

    # Apply any user or group
    if user or group:
        yield chown(dest, user, group, recursive=True)


@operation()
def worktree(
    worktree,
    repo=None, branch=None, create_branch=False, detached=False,
    new_branch=None, commitish=None,
    pull=True, rebase=False, from_remote_branch=None,
    present=True, assume_repo_exists=False, force=False,
    user=None, group=None, state=None, host=None,
):
    '''
    Manage git worktrees.

    + worktree: git working tree directory
    + repo: git main repository directory
    + detached: create a working tree with a detached HEAD
    + branch: (deprecated)
    + create_branch: (deprecated)
    + new_branch: local branch name created at the same time than the worktree
    + commitish: from which git commit, branch, ... the worktree is created
    + pull: pull any changes from a remote branch if set
    + rebase: when pulling, use ``--rebase``
    + from_remote_branch: a 2-tuple ``(remote, branch)`` that identifies a remote branch
    + present: whether the working tree should exist
    + assume_repo_exists: whether to assume the main repo exists
    + force: remove unclean working tree if should not exist
    + user: chown files to this user after
    + group: chown files to this group after

    Example:

    .. code:: python

        git.worktree(
            name='Create a worktree from the current repo `HEAD`',
            repo='/usr/local/src/pyinfra/master',
            worktree='/usr/local/src/pyinfra/hotfix'
        )

        git.worktree(
            name='Create a worktree from the commit `4e091aa0`',
            repo='/usr/local/src/pyinfra/master',
            worktree='/usr/local/src/pyinfra/hotfix',
            commitish='4e091aa0'
        )

        git.worktree(
            name='Create a worktree with a new local branch `v1.0`',
            repo='/usr/local/src/pyinfra/master',
            worktree='/usr/local/src/pyinfra/hotfix',
            new_branch='v1.0',
        )

        git.worktree(
            name='Create a worktree from the commit 4e091aa0 with the new local branch `v1.0`',
            repo='/usr/local/src/pyinfra/master',
            worktree='/usr/local/src/pyinfra/hotfix',
            new_branch='v1.0',
            commitish='4e091aa0'
        )

        git.worktree(
            name='Create a worktree with a detached `HEAD`',
            repo='/usr/local/src/pyinfra/master',
            worktree='/usr/local/src/pyinfra/hotfix',
            detached=True,
        )

        git.worktree(
            name='Create a worktree with a detached `HEAD` from commit `4e091aa0`',
            repo='/usr/local/src/pyinfra/master',
            worktree='/usr/local/src/pyinfra/hotfix',
            commitish='4e091aa0',
            detached=True,
        )

        git.worktree(
            name='Create a worktree from the existing local branch `v1.0`',
            repo='/usr/local/src/pyinfra/master',
            worktree='/usr/local/src/pyinfra/hotfix',
            commitish='v1.0'
        )

        git.worktree(
            name='Create a worktree with a new branch `v1.0` that tracks `origin/v1.0`',
            repo='/usr/local/src/pyinfra/master',
            worktree='/usr/local/src/pyinfra/hotfix',
            new_branch='v1.0',
            commitish='v1.0'
        )

        git.worktree(
            name='Pull an existing worktree already linked to a tracking branch',
            repo='/usr/local/src/pyinfra/master',
            worktree='/usr/local/src/pyinfra/hotfix'
        )

        git.worktree(
            name='Pull an existing worktree from a specific remote branch',
            repo='/usr/local/src/pyinfra/master',
            worktree='/usr/local/src/pyinfra/hotfix',
            from_remote_branch=('origin', 'master')
        )

        git.worktree(
            name='Remove a worktree',
            worktree='/usr/local/src/pyinfra/hotfix',
            present=False,
        )

        git.worktree(
            name='Remove an unclean worktree',
            worktree='/usr/local/src/pyinfra/hotfix',
            present=False,
            force=True,
        )
    '''

    # handle deprecated arguments.
    #  1) old api: branch=xxx, create_branch=True => git worktree add -b xxx <worktree_path>
    #     new api: new_branch=xxx
    #
    #  2) old api: branch=xxx, create_branch=False => git worktree add <worktree_path> xxx
    #     new api: commitish=xxx
    if branch:
        logger.warning(
            'The `branch` and `create_branch` arguments are deprecated. '
            'Please use `branch` and `commitish`.',
        )

        if create_branch:
            if new_branch:
                raise OperationError(
                    'The deprecated arguments `branch` and `create_branch` are not compatible with'
                    ' the new `branch` argument. Please use the new `branch` argument only.',
                )
            else:
                new_branch = branch
        else:
            if commitish:
                raise OperationError(
                    'The deprecated arguments `branch` and `create_branch` are not compatible with'
                    ' the new `commitish` argument. Please use the new `commitish` argument only.',
                )
            else:
                commitish = branch

    # Doesn't exist & we want it
    if not host.fact.directory(worktree) and present:

        # be sure that `repo` is a GIT repository
        if not assume_repo_exists and not host.fact.directory('/'.join((repo, '.git'))):
            raise OperationError(
                'The following folder is not a valid GIT repository : {0}'.format(repo),
            )

        command_parts = ['cd {0} && git worktree add'.format(repo)]

        if new_branch:
            command_parts.append('-b {0}'.format(new_branch))
        elif detached:
            command_parts.append('--detach')

        command_parts.append(worktree)

        if commitish:
            command_parts.append(commitish)

        yield ' '.join(command_parts)

        # Apply any user or group
        if user or group:
            yield chown(worktree, user, group, recursive=True)

    # It exists and we don't want it
    elif host.fact.directory(worktree) and not present:

        command = 'cd {0} && git worktree remove .'.format(worktree)

        if force:
            command += ' --force'

        yield command

    # It exists and we still want it => pull/rebase it
    elif host.fact.directory(worktree) and present:

        # pull the worktree only if it's already linked to a tracking branch or
        # if a remote branch is set
        if host.fact.git_tracking_branch(worktree) or from_remote_branch:
            command = 'cd {0} && git pull'.format(worktree)

            if rebase:
                command += ' --rebase'

            if from_remote_branch:
                if len(from_remote_branch) != 2 or type(from_remote_branch) not in (tuple, list):
                    raise OperationError(
                        'The remote branch must be a 2-tuple (remote, branch) such as '
                        '("origin", "master")',
                    )
                command += ' {0} {1}'.format(*from_remote_branch)

            yield command


@operation
def bare_repo(
    path,
    user=None,
    group=None,
    present=True,
    state=None,
    host=None,
):
    '''
    Create bare git repositories.

    + path: path to the folder
    + present: whether the bare repository should exist
    + user: chown files to this user after
    + group: chown files to this group after

    Example:

    .. code:: python

        git.bare_repo(
            name='Create bare repo',
            path='/home/git/test.git',
        )
    '''

    yield files.directory(path, state=state, host=host, present=present)

    # Ensure our target directory exists
    if present:
        yield 'git init --bare {0}'.format(path)

    # Apply any user or group
    if user or group:
        yield chown(path, user, group, recursive=True)



from .sysvinit import InitdStatus


class RcdStatus(InitdStatus):
    '''
    Same as ``initd_status`` but for BSD (/etc/rc.d) systems. Unlike Linux/init.d,
    BSD init scripts are well behaved and as such their output can be trusted.
    '''

    command = '''
        for SERVICE in `ls /etc/rc.d/`; do
            _=`cat /etc/rc.d/$SERVICE | grep "daemon="`

            if [ "$?" = "0" ]; then
                STATUS=`/etc/rc.d/$SERVICE check`
                echo "$SERVICE=$?"
            fi
        done
    '''

    default = dict



from __future__ import unicode_literals

import click

from pyinfra import logger

from .connectors import EXECUTION_CONNECTORS
from .exceptions import ConnectError, PyinfraError
from .facts import (
    create_host_fact,
    delete_host_fact,
    get_fact_names,
    get_host_fact,
    is_fact,
)


class HostFacts(object):
    def __init__(self, host=None):
        self.host = host

    def __dir__(self):
        return get_fact_names()

    def _check_host(self):
        if not self.host:
            raise TypeError('Cannot call this function with no host!')

    def __getattr__(self, key):
        self._check_host()

        if not is_fact(key):
            raise AttributeError('No such fact: {0}'.format(key))

        # Ensure this host is connected
        connection = self.host.connect(reason='for {0} fact'.format(key))

        # If we can't connect - fail immediately as we specifically need this
        # fact for this host and without it we cannot satisfy the deploy.
        if not connection:
            raise PyinfraError('Could not connect to {0} for fact {1}!'.format(
                self.host, key,
            ))

        return get_host_fact(self.host.state, self.host, key)

    def _create(self, key, data=None, args=None):
        self._check_host()
        return create_host_fact(self.host.state, self.host, key, data, args)

    def _delete(self, key, args=None):
        self._check_host()
        return delete_host_fact(self.host.state, self.host, key, args)


class Host(object):
    '''
    Represents a target host. Thin class that links up to facts and host/group
    data.
    '''

    connection = None
    state = None
    fact = HostFacts()  # this isn't usable, but provides support for dir()

    def __init__(
        self, name, inventory, groups, data,
        executor=EXECUTION_CONNECTORS['ssh'],
    ):
        self.inventory = inventory
        self.groups = groups
        self.data = data
        self.executor = executor
        self.name = name

        # Attach the fact proxy
        self.fact = HostFacts(self)

        # Arbitrary dict for connector use
        self.connector_data = {}

    def __str__(self):
        return '{0}'.format(self.name)

    def __repr__(self):
        return 'Host({0})'.format(self.name)

    @property
    def connected(self):
        return self.connection is not None

    @property
    def host_data(self):
        return self.inventory.get_host_data(self.name)

    @property
    def group_data(self):
        return self.inventory.get_groups_data(self.groups)

    @property
    def print_prefix(self):
        return '{0}[{1}] '.format(
            click.style(''),  # reset
            click.style(self.name, bold=True),
        )

    def style_print_prefix(self, *args, **kwargs):
        return '{0}[{1}] '.format(
            click.style(''),  # reset
            click.style(self.name, *args, **kwargs),
        )

    def noop(self, description):
        '''
        Log a description for a noop operation.
        '''

        handler = logger.info if self.state.print_noop_info else logger.debug
        handler('{0}noop: {1}'.format(self.print_prefix, description))

    # Connector proxy
    #

    def _check_state(self):
        if not self.state:
            raise TypeError('Cannot call this function with no state!')

    def connect(self, reason=None, show_errors=True):
        self._check_state()
        if not self.connection:
            self.state.trigger_callbacks('host_before_connect', self)

            try:
                self.connection = self.executor.connect(self.state, self)
            except ConnectError as e:
                if show_errors:
                    log_message = '{0}{1}'.format(
                        self.print_prefix,
                        click.style(e.args[0], 'red'),
                    )
                    logger.error(log_message)

                self.state.trigger_callbacks('host_connect_error', self, e)
            else:
                log_message = '{0}{1}'.format(
                    self.print_prefix,
                    click.style('Connected', 'green'),
                )
                if reason:
                    log_message = '{0}{1}'.format(
                        log_message,
                        ' ({0})'.format(reason),
                    )

                logger.info(log_message)
                self.state.trigger_callbacks('host_connect', self)

        return self.connection

    def disconnect(self):
        self._check_state()

        # Disconnect is an optional function for executors if needed
        disconnect_func = getattr(self.executor, 'disconnect', None)
        if disconnect_func:
            return disconnect_func(self.state, self)

        self.state.trigger_callbacks('host_disconnect', self)

    def run_shell_command(self, *args, **kwargs):
        self._check_state()
        return self.executor.run_shell_command(self.state, self, *args, **kwargs)

    def put_file(self, *args, **kwargs):
        self._check_state()
        return self.executor.put_file(self.state, self, *args, **kwargs)

    def get_file(self, *args, **kwargs):
        self._check_state()
        return self.executor.get_file(self.state, self, *args, **kwargs)

    # Rsync - optional executor specific ability

    def check_can_rsync(self):
        check_can_rsync_func = getattr(self.executor, 'check_can_rsync', None)
        if check_can_rsync_func:
            return check_can_rsync_func(self)

        raise NotImplementedError('The {0} connector does not support rsync!'.format(
            self.executor.__name__,
        ))

    def rsync(self, *args, **kwargs):
        self._check_state()
        return self.executor.rsync(self.state, self, *args, **kwargs)



# encoding: utf-8

from __future__ import unicode_literals

from unittest import TestCase

from mock import MagicMock, mock_open, patch
from six.moves import shlex_quote

from pyinfra.api import Config, State
from pyinfra.api.connect import connect_all
from pyinfra.api.connectors.util import make_unix_command
from pyinfra.api.exceptions import InventoryError, PyinfraError

from ..util import make_inventory


def fake_ssh_docker_shell(state, host, command,
                          get_pty=False,
                          timeout=None,
                          stdin=None,
                          success_exit_codes=None,
                          print_output=False,
                          print_input=False,
                          return_combined_output=False,
                          use_sudo_password=False,
                          **command_kwargs
                          ):
    if host.data.ssh_hostname not in ('somehost', 'anotherhost'):
        raise PyinfraError('Invalid host', host.data.ssh_hostname)

    if command == 'docker run -d not-an-image tail -f /dev/null':
        return (True, ['containerid'], [])

    elif command == 'docker commit containerid':
        return (True, ['sha256:blahsomerandomstringdata'], [])

    elif command == 'docker rm -f containerid':
        return (True, [], [])

    elif str(command).startswith('rm -f'):
        return (True, [], [])

    # This is a bit kludgy. But it's easier than trying to swap out a mock
    # when it needs to be used...
    elif fake_ssh_docker_shell.custom_command:
        custom_command, status, stdout, stderr = fake_ssh_docker_shell.custom_command
        if str(command) == custom_command:
            fake_ssh_docker_shell.ran_custom_command = True
            return (status, stdout, stderr)

    raise PyinfraError('Invalid Command: {0}'.format(command))


def get_docker_command(command):
    shell_command = make_unix_command(command).get_raw_value()
    shell_command = shlex_quote(shell_command)
    docker_command = 'docker exec -it containerid sh -c {0}'.format(shell_command)
    return docker_command


@patch('pyinfra.api.connectors.ssh.connect', MagicMock())
@patch('pyinfra.api.connectors.ssh.run_shell_command', fake_ssh_docker_shell)
@patch('pyinfra.api.util.open', mock_open(read_data='test!'), create=True)
class TestDockerSSHConnector(TestCase):
    def setUp(self):
        # reset custom command for shell
        fake_ssh_docker_shell.custom_command = None
        fake_ssh_docker_shell.ran_custom_command = False

    def test_connect_host(self):
        inventory = make_inventory(('somehost', 'anotherhost'))
        state = State(inventory, Config())
        host = inventory.get_host('somehost')
        host.connect(reason=True)
        assert len(state.active_hosts) == 0

    def test_missing_image_and_host(self):
        with self.assertRaises(InventoryError):
            make_inventory(hosts=('@dockerssh',))

    def test_missing_image(self):
        with self.assertRaises(InventoryError):
            make_inventory(hosts=('@dockerssh/host',))

        with self.assertRaises(InventoryError):
            make_inventory(hosts=('@dockerssh/host:',))

    def test_connect_all(self):
        inventory = make_inventory(hosts=('@dockerssh/somehost:not-an-image',))
        state = State(inventory, Config())
        connect_all(state)
        assert len(state.active_hosts) == 1

    def test_user_provided_container_id(self):
        inventory = make_inventory(hosts=(
            ('@dockerssh/somehost:not-an-image', {'docker_container_id': 'abc'}),
        ))
        State(inventory, Config())
        host = inventory.get_host('@dockerssh/somehost:not-an-image')
        host.connect()
        assert host.data.docker_container_id == 'abc'

    def test_connect_all_error(self):
        inventory = make_inventory(hosts=('@dockerssh/somehost:a-broken-image',))
        state = State(inventory, Config())

        with self.assertRaises(PyinfraError):
            connect_all(state)

    def test_connect_disconnect_host(self):
        inventory = make_inventory(hosts=('@dockerssh/somehost:not-an-image',))
        state = State(inventory, Config())
        host = inventory.get_host('@dockerssh/somehost:not-an-image')
        host.connect(reason=True)
        assert len(state.active_hosts) == 0
        host.disconnect()

    def test_run_shell_command(self):
        inventory = make_inventory(hosts=('@dockerssh/somehost:not-an-image',))
        State(inventory, Config())

        command = 'echo hi'

        fake_ssh_docker_shell.custom_command = [get_docker_command(command), True, [], []]

        host = inventory.get_host('@dockerssh/somehost:not-an-image')
        host.connect()
        out = host.run_shell_command(
            command,
            stdin='hello',
            get_pty=True,
            print_output=True,
        )
        assert len(out) == 3
        assert out[0] is True
        assert fake_ssh_docker_shell.ran_custom_command

    def test_run_shell_command_error(self):
        inventory = make_inventory(hosts=('@dockerssh/somehost:not-an-image',))
        state = State(inventory, Config())

        command = 'echo hi'
        fake_ssh_docker_shell.custom_command = [get_docker_command(command), False, [], []]

        host = inventory.get_host('@dockerssh/somehost:not-an-image')
        host.connect(state)
        out = host.run_shell_command(command, get_pty=True)
        assert out[0] is False
        assert fake_ssh_docker_shell.ran_custom_command

    @patch('pyinfra.api.connectors.dockerssh.mkstemp', lambda: (None, 'local_tempfile'))
    @patch('pyinfra.api.connectors.docker.os.close', lambda f: None)
    @patch('pyinfra.api.connectors.dockerssh.ssh.put_file')
    def test_put_file(self, fake_put_file):
        fake_ssh_docker_shell.custom_command = [
            'docker cp remote_tempfile containerid:not-another-file', True, [], []]

        inventory = make_inventory(hosts=('@dockerssh/somehost:not-an-image',))
        state = State(inventory, Config())
        state.get_temp_filename = lambda _: 'remote_tempfile'

        host = inventory.get_host('@dockerssh/somehost:not-an-image')
        host.connect()

        host.put_file('not-a-file', 'not-another-file', print_output=True)

        # ensure copy from local to remote host
        fake_put_file.assert_called_with(state, host, 'local_tempfile', 'remote_tempfile')

        # ensure copy from remote host to remote docker container
        assert fake_ssh_docker_shell.ran_custom_command

    @patch('pyinfra.api.connectors.dockerssh.ssh.put_file')
    def test_put_file_error(self, fake_put_file):
        inventory = make_inventory(hosts=('@dockerssh/somehost:not-an-image',))
        state = State(inventory, Config())
        state.get_temp_filename = lambda _: 'remote_tempfile'

        host = inventory.get_host('@dockerssh/somehost:not-an-image')
        host.connect()

        # SSH error
        fake_put_file.return_value = False
        with self.assertRaises(IOError):
            host.put_file('not-a-file', 'not-another-file', print_output=True)

        # docker copy error
        fake_ssh_docker_shell.custom_command = [
            'docker cp remote_tempfile containerid:not-another-file',
            False, [], ['docker error']]
        fake_put_file.return_value = True

        with self.assertRaises(IOError) as exn:
            host.put_file('not-a-file', 'not-another-file', print_output=True)
        assert str(exn.exception) == 'docker error'

    @patch('pyinfra.api.connectors.dockerssh.ssh.get_file')
    def test_get_file(self, fake_get_file):
        fake_ssh_docker_shell.custom_command = [
            'docker cp containerid:not-a-file remote_tempfile', True, [], []]

        inventory = make_inventory(hosts=('@dockerssh/somehost:not-an-image',))
        state = State(inventory, Config())
        state.get_temp_filename = lambda _: 'remote_tempfile'

        host = inventory.get_host('@dockerssh/somehost:not-an-image')
        host.connect()

        host.get_file('not-a-file', 'not-another-file', print_output=True)

        # ensure copy from local to remote host
        fake_get_file.assert_called_with(state, host, 'remote_tempfile', 'not-another-file')

        # ensure copy from remote host to remote docker container
        assert fake_ssh_docker_shell.ran_custom_command

    @patch('pyinfra.api.connectors.dockerssh.ssh.get_file')
    def test_get_file_error(self, fake_get_file):
        fake_ssh_docker_shell.custom_command = [
            'docker cp containerid:not-a-file remote_tempfile', False, [], ['docker error']]

        inventory = make_inventory(hosts=('@dockerssh/somehost:not-an-image',))
        state = State(inventory, Config())
        state.get_temp_filename = lambda _: 'remote_tempfile'

        host = inventory.get_host('@dockerssh/somehost:not-an-image')
        host.connect()

        fake_get_file.return_value = True
        with self.assertRaises(IOError) as ex:
            host.get_file('not-a-file', 'not-another-file', print_output=True)

        assert str(ex.exception) == 'docker error'

        # SSH error
        fake_ssh_docker_shell.custom_command = [
            'docker cp containerid:not-a-file remote_tempfile', True, [], []]
        fake_get_file.return_value = False
        with self.assertRaises(IOError) as ex:
            host.get_file('not-a-file', 'not-another-file', print_output=True)

        assert str(ex.exception) == 'failed to copy file over ssh'



'''
Operations are the core of pyinfra. The ``@operation`` wrapper intercepts calls
to the function and instead diff against the remote server, outputting commands
to the deploy state. This is then run later by pyinfra's ``__main__`` or the
:doc:`./pyinfra.api.operations` module.
'''

from __future__ import unicode_literals

from functools import wraps
from types import FunctionType

import six

import pyinfra

from pyinfra import logger, pseudo_host, pseudo_state

from .command import StringCommand
from .exceptions import OperationValueError, PyinfraError
from .host import Host
from .operation_kwargs import get_execution_kwarg_keys, pop_global_op_kwargs
from .state import State
from .util import (
    get_arg_value,
    get_call_location,
    get_operation_order_from_stack,
    make_hash,
    memoize,
    unroll_generators,
)


# List of available operation names
OPERATIONS = []


def get_operation_names():
    '''
    Returns a list of available operations.
    '''

    return OPERATIONS


class OperationMeta(object):
    def __init__(self, hash=None, commands=None):
        # Wrap all the attributes
        commands = commands or []
        self.commands = commands
        self.hash = hash

        # Changed flag = did we do anything?
        self.changed = len(self.commands) > 0

    def __repr__(self):
        '''Return Operation object as a string.'''
        return 'commands:{} changed:{} hash:{}'.format(
            self.commands, self.changed, self.hash)


def add_op(state, op_func, *args, **kwargs):
    '''
    Prepare & add an operation to ``pyinfra.state`` by executing it on all hosts.

    Args:
        state (``pyinfra.api.State`` obj): the deploy state to add the operation
        to op_func (function): the operation function from one of the modules,
        ie ``server.user``
        args/kwargs: passed to the operation function
    '''

    allow_cli_mode = kwargs.pop('_allow_cli_mode', False)
    after_host_callback = kwargs.pop('_after_host_callback', lambda host: None)

    if pyinfra.is_cli and not allow_cli_mode:
        raise PyinfraError((
            '`add_op` should not be called when pyinfra is executing in CLI mode! ({0})'
        ).format(get_call_location()))

    # This ensures that every time an operation is added (API mode), it is simply
    # appended to the operation order.
    kwargs['_op_order_number'] = len(state.op_meta)

    kwargs['state'] = state

    hosts = kwargs.pop('host', state.inventory.iter_active_hosts())
    if isinstance(hosts, Host):
        hosts = [hosts]

    results = {}
    for host in hosts:
        kwargs['host'] = host
        results[host] = op_func(*args, **kwargs)
        after_host_callback(host)

    return results


@memoize
def show_set_name_warning(call_location):
    logger.warning((
        '{0}:\n\tUse of a `set` to name operations is deprecated, '
        'please us the `name` keyword argument.'
    ).format(call_location))


@memoize
def show_state_host_arguments_warning(call_location):
    logger.warning((
        '{0}:\n\tPassing `state` and `host` as the first two arguments to operations is '
        'deprecated, please use `state` and `host` keyword arguments.'
    ).format(call_location))


def operation(func=None, pipeline_facts=None):
    '''
    Decorator that takes a simple module function and turn it into the internal
    operation representation that consists of a list of commands + options
    (sudo, (sudo|su)_user, env).
    '''

    # If not decorating, return function with config attached
    if func is None:
        def decorator(f):
            setattr(f, 'pipeline_facts', pipeline_facts)
            return operation(f)

        return decorator

    # Index the operation!
    if func.__module__:
        module_bits = func.__module__.split('.')
        module_name = module_bits[-1]
        op_name = '.'.join((module_name, func.__name__))
        OPERATIONS.append(op_name)

    # Actually decorate!
    @wraps(func)
    def decorated_func(*args, **kwargs):
        # Prepare state/host
        #

        # State & host passed in as kwargs (API, nested op, @deploy op)
        if 'state' in kwargs and 'host' in kwargs:
            state = kwargs['state']
            host = kwargs['host']

        # State & host passed in as first two arguments (LEGACY)
        elif len(args) >= 2 and isinstance(args[0], State) and isinstance(args[1], Host):
            show_state_host_arguments_warning(get_call_location())
            state = kwargs['state'] = args[0]
            host = kwargs['host'] = args[1]
            args_copy = list(args)
            args = args_copy[2:]

        # Finally, still no state+host? Use pseudo if we're CLI mode, or fail
        elif pyinfra.is_cli:
            state = kwargs['state'] = pseudo_state._module
            host = kwargs['host'] = pseudo_host._module

            if state.in_op:
                raise PyinfraError((
                    'Nested operation called without state/host: {0} ({1})'
                ).format(op_name, get_call_location()))

            if state.in_deploy:
                raise PyinfraError((
                    'Nested deploy operation called without state/host: {0} ({1})'
                ).format(op_name, get_call_location()))

        else:
            raise PyinfraError((
                'API operation called without state/host: {0} ({1})'
            ).format(op_name, get_call_location()))

        # Configure operation
        #

        # Get the meta kwargs (globals that apply to all hosts)
        global_kwargs, global_kwarg_keys = pop_global_op_kwargs(state, kwargs)

        # If this op is being called inside another, just return here
        # (any unwanted/op-related kwargs removed above).
        if state.in_op:
            if global_kwarg_keys:
                raise PyinfraError((
                    'Nested operation called with global arguments: {0} ({1})'
                ).format(global_kwarg_keys, get_call_location()))
            return func(*args, **kwargs) or []

        # Name the operation
        name = global_kwargs.get('name')
        add_args = False

        if name:
            names = {name}

        # Look for a set as the first argument
        # TODO: remove this! COMPAT w/<1
        elif len(args) > 0 and isinstance(args[0], set):
            show_set_name_warning(get_call_location())
            names = args[0]
            args_copy = list(args)
            args = args[1:]

        # Generate an operation name if needed (Module/Operation format)
        else:
            add_args = True

            if func.__module__:
                module_bits = func.__module__.split('.')
                module_name = module_bits[-1]
                name = '{0}/{1}'.format(module_name.title(), func.__name__.title())
            else:
                name = func.__name__

            names = {name}

        if state.deploy_name:
            names = {
                '{0} | {1}'.format(state.deploy_name, name)
                for name in names
            }

        op_order = get_operation_order_from_stack(state)

        op_order_number = kwargs.pop('_op_order_number', None)
        if op_order_number is not None:
            op_order = [op_order_number]
        # If we're in API mode, op order number must be provided by calling `add_op`
        elif not pyinfra.is_cli:
            raise PyinfraError((
                'Operation order number not provided in API mode - '
                'you must use `add_op` to add operations.'
            ))

        # Make a hash from the call stack lines
        op_hash = make_hash(op_order)

        # Avoid adding duplicates! This happens if an operation is called within
        # a loop - such that the filename/lineno/code _are_ the same, but the
        # arguments might be different. We just append an increasing number to
        # the op hash and also handle below with the op order.
        host_op_hashes = state.meta[host]['op_hashes']
        duplicate_op_count = 0
        while op_hash in host_op_hashes:
            logger.debug('Duplicate hash ({0}) detected!'.format(op_hash))
            op_hash = '{0}-{1}'.format(op_hash, duplicate_op_count)
            duplicate_op_count += 1

        host_op_hashes.add(op_hash)

        if duplicate_op_count:
            op_order.append(duplicate_op_count)

        op_order = tuple(op_order)

        logger.debug('Adding operation, {0}, opOrder={1}, opHash={2}'.format(
            names, op_order, op_hash,
        ))
        state.op_line_numbers_to_hash[op_order] = op_hash

        # Ensure shared (between servers) operation meta
        op_meta = state.op_meta.setdefault(op_hash, {
            'names': set(),
            'args': [],
        })

        for key in get_execution_kwarg_keys():
            global_value = global_kwargs.pop(key)
            op_meta_value = op_meta.get(key)

            if op_meta_value and global_value != op_meta_value:
                raise OperationValueError('Cannot have different values for `{0}`.'.format(key))

            op_meta[key] = global_value

        # Add any meta kwargs (sudo, etc) to the meta - first parse any strings
        # as jinja templates.
        actual_global_kwargs = {
            key: get_arg_value(state, host, a)
            for key, a in six.iteritems(global_kwargs)
        }

        # Add any new names to the set
        op_meta['names'].update(names)

        # Attach normal args, if we're auto-naming this operation
        if add_args:
            for arg in args:
                if isinstance(arg, FunctionType):
                    arg = arg.__name__

                if arg not in op_meta['args']:
                    op_meta['args'].append(arg)

            # Attach keyword args
            for key, value in six.iteritems(kwargs):
                if isinstance(value, FunctionType):
                    value = value.__name__

                if key in ('state', 'host'):
                    continue

                arg = '='.join((str(key), str(value)))
                if arg not in op_meta['args']:
                    op_meta['args'].append(arg)

        # Check if we're actually running the operation on this host
        #

        # Run once and we've already added meta for this op? Stop here.
        if op_meta['run_once']:
            has_run = False
            for ops in six.itervalues(state.ops):
                if op_hash in ops:
                    has_run = True
                    break

            if has_run:
                return OperationMeta(op_hash)

        # "Run" operation
        #

        # Otherwise, flag as in-op and run it to get the commands
        state.in_op = True
        state.current_op_hash = op_hash
        state.current_op_global_kwargs = actual_global_kwargs

        # Generate actual arguments by parsing strings as jinja2 templates. This
        # means you can string format arguments w/o generating multiple
        # operations. Only affects top level operations, as must be run "in_op"
        # so facts are gathered correctly.
        actual_args = [
            get_arg_value(state, host, a)
            for a in args
        ]

        actual_kwargs = {
            key: get_arg_value(state, host, a)
            for key, a in six.iteritems(kwargs)
        }

        # Convert to list as the result may be a generator
        commands = unroll_generators(func(*actual_args, **actual_kwargs))
        commands = [  # convert any strings -> StringCommand's
            StringCommand(command.strip())
            if isinstance(command, six.string_types) else command
            for command in commands
        ]

        state.in_op = False
        state.current_op_hash = None
        state.current_op_global_kwargs = None

        # Add host-specific operation data to state
        #

        # We're doing some commands, meta/ops++
        state.meta[host]['ops'] += 1
        state.meta[host]['commands'] += len(commands)

        # Add the server-relevant commands
        state.ops[host][op_hash] = {
            'commands': commands,
            'global_kwargs': actual_global_kwargs,
        }

        # Return result meta for use in deploy scripts
        return OperationMeta(op_hash, commands)

    decorated_func._pyinfra_op = func
    return decorated_func



from pyinfra.api import FactBase, MaskString, QuoteString, StringCommand
from pyinfra.api.util import try_int

from .util.databases import parse_columns_and_rows


def make_psql_command(
    database=None,
    user=None,
    password=None,
    host=None,
    port=None,
    executable='psql',
):
    target_bits = []

    if password:
        target_bits.append(MaskString('PGPASSWORD="{0}"'.format(password)))

    target_bits.append(executable)

    if database:
        target_bits.append('-d {0}'.format(database))

    if user:
        target_bits.append('-U {0}'.format(user))

    if host:
        target_bits.append('-h {0}'.format(host))

    if port:
        target_bits.append('-p {0}'.format(port))

    return StringCommand(*target_bits)


def make_execute_psql_command(command, **postgresql_kwargs):
    return StringCommand(
        make_psql_command(**postgresql_kwargs),
        '-Ac',
        QuoteString(command),  # quote this whole item as a single shell argument
    )


class PostgresqlFactBase(FactBase):
    abstract = True

    requires_command = 'psql'

    def command(
        self,
        postgresql_user=None, postgresql_password=None,
        postgresql_host=None, postgresql_port=None,
    ):
        return make_execute_psql_command(
            self.postgresql_command,
            user=postgresql_user,
            password=postgresql_password,
            host=postgresql_host,
            port=postgresql_port,
        )


class PostgresqlRoles(PostgresqlFactBase):
    '''
    Returns a dict of PostgreSQL roles and data:

    .. code:: python

        {
            'pyinfra': {
                'super': true,
                'createrole': false,
                'createdb': false,
                ...
            },
        }
    '''

    default = dict
    postgresql_command = 'SELECT * FROM pg_catalog.pg_roles'

    def process(self, output):
        # Remove the last line of the output (row count)
        output = output[:-1]
        rows = parse_columns_and_rows(
            output, '|',
            # Remove the "rol" prefix on column names
            remove_column_prefix='rol',
        )

        users = {}

        for details in rows:
            for key, value in list(details.items()):
                if key in ('oid', 'connlimit'):
                    details[key] = try_int(value)

                if key in (
                    'super', 'inherit', 'createrole', 'createdb',
                    'canlogin', 'replication', 'bypassrls',
                ):
                    details[key] = value == 't'

            users[details.pop('name')] = details

        return users


class PostgresqlDatabases(PostgresqlFactBase):
    '''
    Returns a dict of PostgreSQL databases and metadata:

    .. code:: python

        {
            "pyinfra_stuff": {
                "encoding": "UTF8",
                "collate": "en_US.UTF-8",
                "ctype": "en_US.UTF-8",
                ...
            },
        }
    '''

    default = dict
    postgresql_command = '''
        SELECT pg_catalog.pg_encoding_to_char(encoding), *
        FROM pg_catalog.pg_database
    '''

    def process(self, output):
        # Remove the last line of the output (row count)
        output = output[:-1]
        rows = parse_columns_and_rows(
            output, '|',
            # Remove the "dat" prefix on column names
            remove_column_prefix='dat',
        )

        databases = {}

        for details in rows:
            details['encoding'] = details.pop('pg_encoding_to_char')

            for key, value in list(details.items()):
                if key.endswith('id') or key in (
                    'dba', 'tablespace', 'connlimit',
                ):
                    details[key] = try_int(value)

                if key in ('istemplate', 'allowconn'):
                    details[key] = value == 't'

            databases[details.pop('name')] = details

        return databases



from __future__ import unicode_literals

import re

from pyinfra.api.facts import FactBase


class GitBranch(FactBase):
    requires_command = 'git'

    @staticmethod
    def command(repo):
        return 'cd {0} && git rev-parse --abbrev-ref HEAD'.format(repo)


class GitConfig(FactBase):
    default = dict

    requires_command = 'git'

    @staticmethod
    def command(repo=None):
        if repo is None:
            return 'git config --global -l'

        return 'cd {0} && git config --local -l'.format(repo)

    @staticmethod
    def process(output):
        items = {}

        for line in output:
            key, value = line.split('=', 1)
            items[key] = value

        return items


class GitTrackingBranch(FactBase):
    requires_command = 'git'

    @staticmethod
    def command(repo):
        return r'cd {0} && git status --branch --porcelain'.format(repo)

    @staticmethod
    def process(output):
        if not output:
            return None

        m = re.search(r'\.{3}(\S+)\b', output[0])
        if m:
            return m.group(1)
        return None



from pyinfra.api import FactBase

from .util.packaging import parse_packages

PACMAN_REGEX = r'^([0-9a-zA-Z\-]+)\s([0-9\._+a-z\-]+)'


class PacmanPackages(FactBase):
    '''
    Returns a dict of installed pacman packages:

    .. code:: python

        {
            'package_name': ['version'],
        }
    '''

    command = 'pacman -Q'
    requires_command = 'pacman'

    default = dict

    def process(self, output):
        return parse_packages(PACMAN_REGEX, output)



'''
Manage upstart services.
'''

from __future__ import unicode_literals

from pyinfra.api import operation

from . import files
from .util.service import handle_service_control


@operation
def service(
    service,
    running=True, restarted=False, reloaded=False,
    command=None, enabled=None,
    state=None, host=None,
):
    '''
    Manage the state of upstart managed services.

    + service: name of the service to manage
    + running: whether the service should be running
    + restarted: whether the service should be restarted
    + reloaded: whether the service should be reloaded
    + command: custom command to pass like: ``/etc/rc.d/<service> <command>``
    + enabled: whether this service should be enabled/disabled on boot

    Enabling/disabling services:
        Upstart jobs define runlevels in their config files - as such there is no way to
        edit/list these without fiddling with the config. So pyinfra simply manages the
        existence of a ``/etc/init/<service>.override`` file, and sets its content to
        "manual" to disable automatic start of services.
    '''

    yield handle_service_control(
        host,
        service, host.fact.upstart_status,
        'initctl {1} {0}',
        running, restarted, reloaded, command,
    )

    # Upstart jobs are setup w/runlevels etc in their config files, so here we just check
    # there's no override file.
    if enabled is True:
        yield files.file(
            '/etc/init/{0}.override'.format(service),
            present=False,
            state=state, host=host,
        )

    # Set the override file to "manual" to disable automatic start
    elif enabled is False:
        yield 'echo "manual" > /etc/init/{0}.override'.format(service)



from __future__ import print_function, unicode_literals

import click
import winrm

from pyinfra import logger
from pyinfra.api import Config
from pyinfra.api.exceptions import ConnectError, PyinfraError
from pyinfra.api.util import memoize

from .util import make_win_command


def _raise_connect_error(host, message, data):
    message = '{0} ({1})'.format(message, data)
    raise ConnectError(message)


@memoize
def show_warning():
    logger.warning('The @winrm connector is alpha!')


def _make_winrm_kwargs(state, host):
    kwargs = {
    }

    for key, value in (
        ('username', host.data.winrm_user),
        ('password', host.data.winrm_password),
        ('winrm_port', int(host.data.winrm_port or 0)),
        ('timeout', state.config.CONNECT_TIMEOUT),
    ):
        if value:
            kwargs[key] = value

    # Password auth (boo!)
    if host.data.winrm_password:
        kwargs['password'] = host.data.winrm_password

    # FUTURE: add more auth
    # pywinrm supports: basic, certificate, ntlm, kerberos, plaintext, ssl, credssp
    # see https://github.com/diyan/pywinrm/blob/master/winrm/__init__.py#L12

    return kwargs


def make_names_data(hostname):

    show_warning()

    yield '@winrm/{0}'.format(hostname), {'winrm_hostname': hostname}, []


def connect(state, host):
    '''
    Connect to a single host. Returns the winrm Session if successful.
    '''

    kwargs = _make_winrm_kwargs(state, host)
    logger.debug('Connecting to: {0} ({1})'.format(host.name, kwargs))

    # Hostname can be provided via winrm config (alias), data, or the hosts name
    hostname = kwargs.pop(
        'hostname',
        host.data.winrm_hostname or host.name,
    )

    logger.debug('winrm_username:{} winrm_password:{} '
                 'winrm_port:{}'.format(host.data.winrm_username, host.data.winrm_password,
                                        host.data.winrm_port))

    try:
        # Create new session
        host_and_port = '{}:{}'.format(hostname, host.data.winrm_port)
        logger.debug('host_and_port: {}'.format(host_and_port))

        session = winrm.Session(
            host_and_port,
            auth=(
                host.data.winrm_username,
                host.data.winrm_password,
            ),
        )

        return session

    # TODO: add exceptions here
    except Exception as e:
        auth_kwargs = {}

        for key, value in kwargs.items():
            if key in ('username', 'password'):
                auth_kwargs[key] = value
                continue

        auth_args = ', '.join(
            '{0}={1}'.format(key, value)
            for key, value in auth_kwargs.items()
        )
        logger.debug(str(e))
        _raise_connect_error(host, 'Authentication error', auth_args)


def run_shell_command(
    state, host, command,
    get_pty=False,
    sudo=None,
    sudo_user=None,
    su_user=None,
    use_sudo_login=False,
    use_su_login=False,
    preserve_sudo_env=False,
    timeout=None,
    stdin=None,
    success_exit_codes=None,
    print_output=False,
    print_input=False,
    return_combined_output=False,
    shell_executable=Config.SHELL,
    **command_kwargs
):
    '''
    Execute a command on the specified host.

    Args:
        state (``pyinfra.api.State`` obj): state object for this command
        hostname (string): hostname of the target
        command (string): actual command to execute
        get_pty (boolean): Not used for WINRM
        sudo (boolean): Not used for WINRM
        sudo_user (string): Not used for WINRM
        use_sudo_login(boolean): Not used for WINRM
        use_su_login(boolean): Not used for WINRM
        preserve_sudo_env(boolean): Not used for WINRM
    TODO: check if winrm has a timeout and use timeout param
        timeout (int): timeout for this command to complete before erroring
        stdin (string): Not used for WINRM
        success_exit_codes (list): all values in the list that will return success
        print_output (boolean): print the output
        TODO print_intput (boolean): print the input
        return_combined_output (boolean): combine the stdout and stderr lists
        shell_executable (string): shell to use - 'sh'=cmd, 'ps'=powershell(default)
        env (dict): environment variables to set

    Returns:
        tuple: (exit_code, stdout, stderr)
        stdout and stderr are both lists of strings from each buffer.
    '''

    command = make_win_command(command, **command_kwargs)

    logger.debug('Running command on {0}: {1}'.format(host.name, command))

    # get rid of leading/trailing quote
    tmp_command = command.strip("'")

    if print_output:
        click.echo(
            '{0}>>> {1}'.format(host.print_prefix, command),
            err=True,
        )

    if not shell_executable:
        shell_executable = 'ps'
    logger.debug('shell_executable:{0}'.format(shell_executable))

    # default windows to use ps, but allow it to be overridden
    if shell_executable in ['cmd']:
        response = host.connection.run_cmd(tmp_command)
    else:
        response = host.connection.run_ps(tmp_command)

    return_code = response.status_code
    logger.debug('response:{}'.format(response))

    std_out_str = response.std_out.decode('utf-8')
    std_err_str = response.std_err.decode('utf-8')

    # split on '\r\n' (windows newlines)
    std_out = std_out_str.split('\r\n')
    std_err = std_err_str.split('\r\n')

    logger.debug('std_out:{}'.format(std_out))
    logger.debug('std_err:{}'.format(std_err))

    if print_output:
        click.echo(
            '{0}>>> {1}'.format(host.print_prefix, '\n'.join(std_out)),
            err=True,
        )

    if success_exit_codes:
        status = return_code in success_exit_codes
    else:
        status = return_code == 0

    logger.debug('Command exit status: {0}'.format(status))

    if return_combined_output:
        return status, std_out + std_err

    return status, std_out, std_err


def get_file(
    state, host, remote_filename, filename_or_io,
    **command_kwargs
):
    raise PyinfraError('Not implemented')


def put_file(
    state, host, filename_or_io, remote_filename,
    **command_kwargs
):
    raise PyinfraError('Not implemented')


EXECUTION_CONNECTOR = True



'''
Manage MySQL databases, users and privileges.

Requires the ``mysql`` CLI executable on the target host(s).

All operations in this module take four optional global arguments:
    + ``mysql_user``: the username to connect to mysql to
    + ``mysql_password``: the password for the connecting user
    + ``mysql_host``: the hostname of the server to connect to
    + ``mysql_port``: the port of the server to connect to

See the example/mysql.py

'''

import six

from pyinfra.api import MaskString, operation, OperationError, StringCommand
from pyinfra.facts.mysql import make_execute_mysql_command, make_mysql_command


@operation
def sql(
    sql,
    database=None,
    # Details for speaking to MySQL via `mysql` CLI
    mysql_user=None, mysql_password=None,
    mysql_host=None, mysql_port=None,
    state=None, host=None,
):
    '''
    Execute arbitrary SQL against MySQL.

    + sql: SQL command(s) to execute
    + database: optional database to open the connection with
    + mysql_*: global module arguments, see above
    '''

    yield make_execute_mysql_command(
        sql,
        database=database,
        user=mysql_user,
        password=mysql_password,
        host=mysql_host,
        port=mysql_port,
    )


@operation
def user(
    user,
    # Desired user settings
    present=True,
    user_hostname='localhost', password=None, privileges=None,
    # Details for speaking to MySQL via `mysql` CLI via `mysql` CLI
    mysql_user=None, mysql_password=None,
    mysql_host=None, mysql_port=None,
    state=None, host=None,
):
    '''
    Add/remove/update MySQL users.

    + user: the name of the user
    + present: whether the user should exist or not
    + user_hostname: the hostname of the user
    + password: the password of the user (if created)
    + privileges: the global privileges for this user
    + mysql_*: global module arguments, see above

    Hostname:
        this + ``name`` makes the username - so changing this will create a new
        user, rather than update users with the same ``name``.

    Password:
        will only be applied if the user does not exist - ie pyinfra cannot
        detect if the current password doesn't match the one provided, so won't
        attempt to change it.

    Example:

    .. code:: python

        mysql.user(
            name='Create the pyinfra@localhost MySQL user',
            user='pyinfra',
            password='somepassword',
        )
    '''

    current_users = host.fact.mysql_users(
        mysql_user, mysql_password, mysql_host, mysql_port,
    )

    user_host = '{0}@{1}'.format(user, user_hostname)
    is_present = user_host in current_users

    # User not wanted?
    if not present:
        if is_present:
            yield make_execute_mysql_command(
                'DROP USER "{0}"@"{1}"'.format(user, user_hostname),
                user=mysql_user,
                password=mysql_password,
                host=mysql_host,
                port=mysql_port,
            )
        else:
            host.noop('mysql user {0}@{1} does not exist'.format(user, user_hostname))
        return

    # If we want the user and they don't exist
    if present and not is_present:
        sql_bits = ['CREATE USER "{0}"@"{1}"'.format(user, user_hostname)]
        if password:
            sql_bits.append(MaskString('IDENTIFIED BY "{0}"'.format(password)))

        yield make_execute_mysql_command(
            StringCommand(*sql_bits),
            user=mysql_user,
            password=mysql_password,
            host=mysql_host,
            port=mysql_port,
        )
    else:
        host.noop('mysql user {0}@{1} exists'.format(user, user_hostname))

    # If we're here either the user exists or we just created them; either way
    # now we can check any privileges are set.
    if privileges:
        yield _privileges(
            user, privileges,
            user_hostname=user_hostname,
            mysql_user=mysql_user, mysql_password=mysql_password,
            mysql_host=mysql_host, mysql_port=mysql_port,
            state=state, host=host,
        )


@operation
def database(
    database,
    # Desired database settings
    present=True,
    collate=None, charset=None,
    user=None, user_hostname='localhost', user_privileges='ALL',
    # Details for speaking to MySQL via `mysql` CLI
    mysql_user=None, mysql_password=None,
    mysql_host=None, mysql_port=None,
    state=None, host=None,
):
    '''
    Add/remove MySQL databases.

    + name: the name of the database
    + present: whether the database should exist or not
    + collate: the collate to use when creating the database
    + charset: the charset to use when creating the database
    + user: MySQL user to grant privileges on this database to
    + user_hostname: the hostname of the MySQL user to grant
    + user_privileges: privileges to grant to any specified user
    + mysql_*: global module arguments, see above

    Collate/charset:
        these will only be applied if the database does not exist - ie pyinfra
        will not attempt to alter the existing databases collate/character sets.

    Example:

    .. code:: python

        mysql.database(
            name='Create the pyinfra_stuff database',
            database='pyinfra_stuff',
            user='pyinfra',
            user_privileges=['SELECT', 'INSERT'],
            charset='utf8',
        )
    '''

    current_databases = host.fact.mysql_databases(
        mysql_user, mysql_password,
        mysql_host, mysql_port,
    )

    is_present = database in current_databases

    if not present:
        if is_present:
            yield make_execute_mysql_command(
                'DROP DATABASE {0}'.format(database),
                user=mysql_user,
                password=mysql_password,
                host=mysql_host,
                port=mysql_port,
            )
        else:
            host.noop('mysql database {0} does not exist'.format(database))
        return

    # We want the database but it doesn't exist
    if present and not is_present:
        sql_bits = ['CREATE DATABASE {0}'.format(database)]

        if collate:
            sql_bits.append('COLLATE {0}'.format(collate))

        if charset:
            sql_bits.append('CHARSET {0}'.format(charset))

        yield make_execute_mysql_command(
            ' '.join(sql_bits),
            user=mysql_user,
            password=mysql_password,
            host=mysql_host,
            port=mysql_port,
        )
    else:
        host.noop('mysql database {0} exists'.format(database))

    # Ensure any user privileges for this database
    if user and user_privileges:
        yield privileges(
            user,
            user_hostname=user_hostname,
            privileges=user_privileges,
            database=database,
            state=state,
            host=host,
        )


@operation
def privileges(
    user, privileges,
    user_hostname='localhost',
    database='*', table='*',
    present=True,
    flush=True,
    # Details for speaking to MySQL via `mysql` CLI
    mysql_user=None, mysql_password=None,
    mysql_host=None, mysql_port=None,
    state=None, host=None,
):
    '''
    Add/remove MySQL privileges for a user, either global, database or table specific.

    + user: name of the user to manage privileges for
    + privileges: list of privileges the user should have
    + user_hostname: the hostname of the user
    + database: name of the database to grant privileges to (defaults to all)
    + table: name of the table to grant privileges to (defaults to all)
    + present: whether these privileges should exist (False to ``REVOKE)
    + flush: whether to flush (and update) the privileges table after any changes
    + mysql_*: global module arguments, see above
    '''

    # Ensure we have a list
    if isinstance(privileges, six.string_types):
        privileges = [privileges]

    if database != '*':
        database = '`{0}`'.format(database)

    if table != '*':
        table = '`{0}`'.format(table)

        # We can't set privileges on *.tablename as MySQL won't allow it
        if database == '*':
            raise OperationError((
                'Cannot apply MySQL privileges on {0}.{1}, no database provided'
            ).format(database, table))

    database_table = '{0}.{1}'.format(database, table)
    user_grants = host.fact.mysql_user_grants(
        user, user_hostname,
        mysql_user, mysql_password,
        mysql_host, mysql_port,
    )

    has_privileges = False

    if database_table in user_grants:
        existing_privileges = [
            'ALL' if privilege == 'ALL PRIVILEGES' else privilege
            for privilege in user_grants[database_table]['privileges']
        ]

        has_privileges = (
            database_table in user_grants
            and all(
                privilege in existing_privileges
                for privilege in privileges
            )
        )

    target = action = None

    # No privilege and we want it
    if present:
        if not has_privileges:
            action = 'GRANT'
            target = 'TO'
        else:
            host.noop('mysql privileges exist')
            return

    # Permission we don't want
    if not present:
        if has_privileges:
            action = 'REVOKE'
            target = 'FROM'
        else:
            host.noop('mysql privileges do not exist')
            return

    if target and action:
        command = (
            '{action} {privileges} '
            'ON {database}.{table} '
            '{target} "{user}"@"{user_hostname}"'
        ).format(
            privileges=', '.join(privileges),
            action=action, target=target,
            database=database, table=table,
            user=user, user_hostname=user_hostname,
        ).replace('`', r'\`')

        yield make_execute_mysql_command(
            command,
            user=mysql_user,
            password=mysql_password,
            host=mysql_host,
            port=mysql_port,
        )

        if flush:
            yield make_execute_mysql_command(
                'FLUSH PRIVILEGES',
                user=mysql_user,
                password=mysql_password,
                host=mysql_host,
                port=mysql_port,
            )

_privileges = privileges  # noqa: E305 (for use where kwarg is the same)


@operation
def dump(
    dest, database=None,
    # Details for speaking to MySQL via `mysql` CLI
    mysql_user=None, mysql_password=None,
    mysql_host=None, mysql_port=None,
    state=None, host=None,
):
    '''
    Dump a MySQL database into a ``.sql`` file. Requires ``mysqldump``.

    + dest: name of the file to dump the SQL to
    + database: name of the database to dump
    + mysql_*: global module arguments, see above

    Example:

    .. code:: python

        mysql.dump(
            name='Dump the pyinfra_stuff database',
            dest='/tmp/pyinfra_stuff.dump',
            database='pyinfra_stuff',
        )
    '''

    yield '{0} > {1}'.format(make_mysql_command(
        executable='mysqldump',
        database=database,
        user=mysql_user,
        password=mysql_password,
        host=mysql_host,
        port=mysql_port,
    ), dest)


@operation
def load(
    src, database=None,
    # Details for speaking to MySQL via `mysql` CLI
    mysql_user=None, mysql_password=None,
    mysql_host=None, mysql_port=None,
    state=None, host=None,
):
    '''
    Load ``.sql`` file into a database.

    + src: the filename to read from
    + database: name of the database to import into
    + mysql_*: global module arguments, see above

    Example:

    .. code:: python

        mysql.load(
            name='Import the pyinfra_stuff dump into pyinfra_stuff_copy',
            src='/tmp/pyinfra_stuff.dump',
            database='pyinfra_stuff_copy',
        )
    '''

    yield '{0} < {1}'.format(make_mysql_command(
        database=database,
        user=mysql_user,
        password=mysql_password,
        host=mysql_host,
        port=mysql_port,
    ), src)



'''
These three pseudo modules (state, inventory, host) are used throughout pyinfra
and provide the magic that means "from pyinfra import host" inside a deploy
file always represents the *current* host being executed, ie these modules are
dynamic and change during execution of pyinfra.

Although CLI only when in use, these are bundled into the main pyinfra package
as they are utilised throughout (to determine the current state/host when
executing in CLI mode).
'''

import sys

import pyinfra


class PseudoModule(object):
    _module = None
    _base_module = None

    def __repr__(self):
        return 'PseudoModule({0}):{1}'.format(
            self._base_module.__name__,
            repr(self._module),
        )

    def __str__(self):
        return str(self._module)

    def __dir__(self):
        return dir(self._base_module)

    def __getattr__(self, key):
        if self._module is None:
            return getattr(self._base_module, key)
        return getattr(self._module, key)

    def __iter__(self):
        return iter(self._module)

    def __len__(self):
        return len(self._module)

    def __eq__(self, other):
        return self._module == other

    def set(self, module):
        self._module = module

    def set_base(self, module):
        self._base_module = module

    def reset(self):
        self._module = None

    def isset(self):
        return self._module is not None


# The current deploy state
pseudo_state = \
    sys.modules['pyinfra.pseudo_state'] = sys.modules['pyinfra.state'] = \
    pyinfra.pseudo_state = pyinfra.state = \
    PseudoModule()

# The current deploy inventory
pseudo_inventory = \
    sys.modules['pyinfra.pseudo_inventory'] = sys.modules['pyinfra.inventory'] = \
    pyinfra.pseudo_inventory = pyinfra.inventory = \
    PseudoModule()

# The current target host
pseudo_host = \
    sys.modules['pyinfra.pseudo_host'] = sys.modules['pyinfra.host'] = \
    pyinfra.pseudo_host = pyinfra.host = \
    PseudoModule()


def init_base_classes():
    from pyinfra.api import Host, Inventory, State

    pseudo_host.set_base(Host)
    pseudo_inventory.set_base(Inventory)
    pseudo_state.set_base(State)



from __future__ import division, unicode_literals

from contextlib import contextmanager
from multiprocessing import cpu_count
from os import path
from uuid import uuid4

import six

from gevent.pool import Pool
from pkg_resources import parse_version, require, Requirement, ResolutionError

from pyinfra import __version__, logger

from .config import Config
from .exceptions import PyinfraError
from .util import get_caller_frameinfo, sha1_hash

# Work out the max parallel we can achieve with the open files limit of the user/process,
# take 10 for opening Python files and /3 for ~3 files per host during op runs.
# See: https://github.com/Fizzadar/pyinfra/issues/44
try:
    from resource import getrlimit, RLIMIT_NOFILE
    nofile_limit, _ = getrlimit(RLIMIT_NOFILE)
    MAX_PARALLEL = round((nofile_limit - 10) / 3)

# Resource isn't available on Windows
except ImportError:
    nofile_limit = 0
    MAX_PARALLEL = None


class BaseStateCallback(object):
    # Host callbacks
    #

    @staticmethod
    def host_before_connect(state, host):
        pass

    @staticmethod
    def host_connect(state, host):
        pass

    @staticmethod
    def host_connect_error(state, host, error):
        pass

    @staticmethod
    def host_disconnect(state, host):
        pass

    # Operation callbacks
    #

    @staticmethod
    def operation_start(state, op_hash):
        pass

    @staticmethod
    def operation_host_start(state, host, op_hash):
        pass

    @staticmethod
    def operation_host_success(state, host, op_hash):
        pass

    @staticmethod
    def operation_host_error(state, host, op_hash):
        pass

    @staticmethod
    def operation_end(state, op_hash):
        pass


class State(object):
    '''
    Manages state for a pyinfra deploy.
    '''

    initialised = False

    # A pyinfra.api.Inventory which stores all our pyinfra.api.Host's
    inventory = None

    # A pyinfra.api.Config
    config = None

    # Main gevent pool
    pool = None

    # Whether we are in an @operation (so inner ops aren't wrapped)
    in_op = False
    # Whether we are deploying (ie hosts are all ready)
    deploying = False

    # Current op hash
    current_op_hash = None
    # Current op global args for use w/facts
    current_op_global_kwargs = None

    loop_counter = None
    loop_line = None
    loop_filename = None

    # Name of the current deploy
    in_deploy = False
    deploy_name = None
    deploy_kwargs = None
    deploy_data = None
    deploy_line_numbers = None

    print_noop_info = False  # print "[host] noop: reason for noop"
    print_fact_info = False  # print "loaded fact X"
    print_input = False
    print_fact_input = False
    print_output = False
    print_fact_output = False

    # Used in CLI
    deploy_dir = None  # base directory for locating files/templates/etc
    current_op_file = 0  # increments for each file excuted to maintain order
    current_deploy_filename = None

    def __init__(self, inventory=None, config=None, **kwargs):
        if inventory:
            self.init(inventory, config, **kwargs)

    def init(self, inventory, config, initial_limit=None):
        # Config validation
        #

        # If no config, create one using the defaults
        if config is None:
            config = Config()

        # Error if our min version is not met
        if config.MIN_PYINFRA_VERSION is not None:
            # TODO: remove this
            if config.REQUIRE_PYINFRA_VERSION is None:
                config.REQUIRE_PYINFRA_VERSION = '>={0}'.format(config.MIN_PYINFRA_VERSION)
                logger.warning(
                    '`MIN_PYINFRA_VERSION` is deprecated, please use `REQUIRE_PYINFRA_VERSION`.',
                )
            else:
                logger.warning(
                    'Ignoring legacy `MIN_PYINFRA_VERSION` because '
                    '`REQUIRE_PYINFRA_VERSION` also exists.',
                )

        if config.REQUIRE_PYINFRA_VERSION is not None:
            running_version = parse_version(__version__)
            required_versions = Requirement.parse(
                'pyinfra{0}'.format(config.REQUIRE_PYINFRA_VERSION),
            )

            if running_version not in required_versions:
                raise PyinfraError((
                    'pyinfra version requirement not met '
                    '(requires {0}, running {1})'
                ).format(
                    config.REQUIRE_PYINFRA_VERSION,
                    __version__,
                ))

        if config.REQUIRE_PACKAGES is not None:
            if isinstance(config.REQUIRE_PACKAGES, (list, tuple)):
                requirements = config.REQUIRE_PACKAGES
            else:
                with open(path.join(self.deploy_dir, config.REQUIRE_PACKAGES)) as f:
                    requirements = [
                        line.split('#egg=')[-1]
                        for line in f.read().splitlines()
                    ]

            try:
                require(requirements)
            except ResolutionError as e:
                raise PyinfraError('Deploy requirements ({0}) not met: {1}'.format(
                    config.REQUIRE_PACKAGES, e,
                ))

        if not config.PARALLEL:
            # TODO: benchmark this
            # In my own tests the optimum number of parallel SSH processes is
            # ~20 per CPU core - no science here yet, needs benchmarking!
            cpus = cpu_count()
            ideal_parallel = cpus * 20

            config.PARALLEL = (
                min(ideal_parallel, len(inventory), MAX_PARALLEL)
                if MAX_PARALLEL is not None
                else min(ideal_parallel, len(inventory))
            )

        # If explicitly set, just issue a warning
        elif MAX_PARALLEL is not None and config.PARALLEL > MAX_PARALLEL:
            logger.warning((
                'Parallel set to {0}, but this may hit the open files limit of {1}.\n'
                '    Max recommended value: {2}'
            ).format(config.PARALLEL, nofile_limit, MAX_PARALLEL))

        # Actually initialise the state object
        #

        self.callback_handlers = []

        # Setup greenlet pools
        self.pool = Pool(config.PARALLEL)
        self.fact_pool = Pool(config.PARALLEL)

        # Connection storage
        self.ssh_connections = {}
        self.sftp_connections = {}

        # Private keys
        self.private_keys = {}

        # Assign inventory/config
        self.inventory = inventory
        self.config = config

        # Hosts we've activated at any time
        self.activated_hosts = set()
        # Active hosts that *haven't* failed yet
        self.active_hosts = set()
        # Hosts that have failed
        self.failed_hosts = set()

        # Limit hosts changes dynamically to limit operations to a subset of hosts
        self.limit_hosts = initial_limit

        # Op basics
        self.op_line_numbers_to_hash = {}
        self.op_meta = {}  # maps operation hash -> names/etc
        self.ops_run = set()  # list of ops which have been started/run

        # Op dict for each host
        self.ops = {
            host: {}
            for host in inventory
        }

        # Facts dict for each host
        self.facts = {
            host: {}
            for host in inventory
        }

        # Meta dict for each host
        self.meta = {
            host: {
                'ops': 0,  # one function call in a deploy file
                'commands': 0,  # actual # of commands to run
                'op_hashes': set(),
            }
            for host in inventory
        }

        # Results dict for each host
        self.results = {
            host: {
                'ops': 0,  # success_ops + failed ops w/ignore_errors
                'success_ops': 0,
                'error_ops': 0,
                'commands': 0,
            }
            for host in inventory
        }

        # Assign state back references to inventory & config
        inventory.state = config.state = self
        for host in inventory:
            host.state = self

        self.initialised = True

        # Flag to track added users (via `server.user` operation calls). This is
        # specifically to address users not existing during fact gathering phase
        # causing failures with su_user/sudo_user. If we expect to add the user
        # those facts should not fail but default.
        self.will_add_users = []

    def will_add_user(self, username):
        return username in self.will_add_users

    def add_will_add_user(self, username):
        self.will_add_users.append(username)

    def to_dict(self):
        return {
            'op_order': self.get_op_order(),
            'ops': self.ops,
            'facts': self.facts,
            'meta': self.meta,
            'results': self.results,
        }

    def add_callback_handler(self, handler):
        if not isinstance(handler, BaseStateCallback):
            raise TypeError((
                '{0} is not a valid callback handler (use `BaseStateCallback`)'
            ).format(handler))
        self.callback_handlers.append(handler)

    def trigger_callbacks(self, method_name, *args, **kwargs):
        for handler in self.callback_handlers:
            func = getattr(handler, method_name)
            func(self, *args, **kwargs)

    @contextmanager
    def deploy(self, name, kwargs, data, in_deploy=True):
        '''
        Wraps a group of operations as a deploy, this should not be used
        directly, instead use ``pyinfra.api.deploy.deploy``.
        '''

        # Handle nested deploy names
        if self.deploy_name:
            name = '{0} | {1}'.format(self.deploy_name, name)

        # Store the previous values
        old_in_deploy = self.in_deploy
        old_deploy_name = self.deploy_name
        old_deploy_kwargs = self.deploy_kwargs
        old_deploy_data = self.deploy_data
        self.in_deploy = in_deploy

        # Set the new values
        self.deploy_name = name
        self.deploy_kwargs = kwargs
        self.deploy_data = data
        logger.debug('Starting deploy {0} (args={1}, data={2})'.format(
            name, kwargs, data,
        ))

        yield

        # Restore the previous values
        self.in_deploy = old_in_deploy
        self.deploy_name = old_deploy_name
        self.deploy_kwargs = old_deploy_kwargs
        self.deploy_data = old_deploy_data

        logger.debug('Reset deploy to {0} (args={1}, data={2})'.format(
            old_deploy_name, old_deploy_kwargs, old_deploy_data,
        ))

    @contextmanager
    def preserve_loop_order(self, items):
        frameinfo = get_caller_frameinfo(frame_offset=1)  # escape contextlib
        self.loop_line = frameinfo.lineno
        self.loop_filename = frameinfo.filename

        def item_generator():
            for i, item in enumerate(items, 1):
                self.loop_counter = i
                yield item

        yield item_generator

        self.loop_counter = None
        self.loop_line = None
        self.loop_filename = None

    def get_op_order(self):
        line_numbers_to_hash = self.op_line_numbers_to_hash
        sorted_line_numbers = sorted(list(line_numbers_to_hash.keys()))

        return [
            line_numbers_to_hash[numbers]
            for numbers in sorted_line_numbers
        ]

    def get_op_meta(self, op_hash):
        return self.op_meta[op_hash]

    def get_op_data(self, host, op_hash):
        return self.ops[host][op_hash]

    def activate_host(self, host):
        '''
        Flag a host as active.
        '''

        logger.debug('Activating host: {0}'.format(host))

        # Add to *both* activated and active - active will reduce as hosts fail
        # but connected will not, enabling us to track failed %.
        self.activated_hosts.add(host)
        self.active_hosts.add(host)

    def fail_hosts(self, hosts_to_fail, activated_count=None):
        '''
        Flag a ``set`` of hosts as failed, error for ``config.FAIL_PERCENT``.
        '''

        if not hosts_to_fail:
            return

        activated_count = activated_count or len(self.activated_hosts)

        logger.debug('Failing hosts: {0}'.format(', '.join(
            (host.name for host in hosts_to_fail),
        )))

        self.failed_hosts.update(hosts_to_fail)

        self.active_hosts -= hosts_to_fail

        # Check we're not above the fail percent
        active_hosts = self.active_hosts

        # No hosts left!
        if not active_hosts:
            raise PyinfraError('No hosts remaining!')

        if self.config.FAIL_PERCENT is not None:
            percent_failed = (
                1 - len(active_hosts) / activated_count
            ) * 100

            if percent_failed > self.config.FAIL_PERCENT:
                raise PyinfraError('Over {0}% of hosts failed ({1}%)'.format(
                    self.config.FAIL_PERCENT,
                    int(round(percent_failed)),
                ))

    def is_host_in_limit(self, host):
        '''
        Returns a boolean indicating if the host is within the current state limit.
        '''

        limit_hosts = self.limit_hosts

        if not isinstance(limit_hosts, list):
            return True
        return host in limit_hosts

    def get_temp_filename(self, hash_key=None, hash_filename=True):
        '''
        Generate a temporary filename for this deploy.
        '''

        if not hash_key:
            hash_key = six.text_type(uuid4())

        if hash_filename:
            hash_key = sha1_hash(hash_key)

        return '{0}/pyinfra-{1}'.format(self.config.TEMP_DIR, hash_key)



from datetime import datetime
from unittest import TestCase

from six.moves import cStringIO as StringIO

from pyinfra.operations import server
from pyinfra_cli.exceptions import CliError
from pyinfra_cli.util import get_operation_and_args, json_encode


class TestCliUtil(TestCase):
    def test_json_encode_function(self):
        assert json_encode(get_operation_and_args) == 'Function: get_operation_and_args'

    def test_json_encode_datetime(self):
        now = datetime.utcnow()
        assert json_encode(now) == now.isoformat()

    def test_json_encode_file(self):
        file = StringIO()
        assert json_encode(file) == 'In memory file: '

    def test_json_encode_set(self):
        assert json_encode({1, 2, 3}) == [1, 2, 3]

    def test_setup_no_module(self):
        with self.assertRaises(CliError) as context:
            get_operation_and_args(('no.op',))
        assert context.exception.message == 'No such module: no'

    def test_setup_no_op(self):
        with self.assertRaises(CliError) as context:
            get_operation_and_args(('server.no',))

        assert context.exception.message == 'No such operation: server.no'

    def test_setup_op_and_args(self):
        commands = ('server.user', 'one', 'two', 'hello=world')

        assert get_operation_and_args(commands) == (
            server.user,
            (['one', 'two'], {'hello': 'world'}),
        )

    def test_setup_op_and_json_args(self):
        commands = ('server.user', '[["one", "two"], {"hello": "world"}]')

        assert get_operation_and_args(commands) == (
            server.user,
            (['one', 'two'], {'hello': 'world'}),
        )



from __future__ import unicode_literals

import os
import re
import shutil
from datetime import datetime
from tempfile import mkdtemp

from dateutil.parser import parse as parse_date

from pyinfra.api import FactBase, ShortFactBase
from pyinfra.api.util import try_int

from .util.distro import get_distro_info


class User(FactBase):
    '''
    Returns the name of the current user.
    '''

    command = 'echo $USER'


class Home(FactBase):
    '''
    Returns the home directory of the current user.
    '''

    command = 'echo $HOME'


class Hostname(FactBase):
    '''
    Returns the current hostname of the server.
    '''

    command = 'uname -n'


class Os(FactBase):
    '''
    Returns the OS name according to ``uname``.
    '''

    command = 'uname -s'


class OsVersion(FactBase):
    '''
    Returns the OS version according to ``uname``.
    '''

    command = 'uname -r'


class Arch(FactBase):
    '''
    Returns the system architecture according to ``uname``.
    '''

    command = 'uname -p'


class Command(FactBase):
    '''
    Returns the raw output lines of a given command.
    '''

    @staticmethod
    def command(command):
        return command


class Which(FactBase):
    '''
    Returns the path of a given command, if available.
    '''

    @staticmethod
    def command(name):
        return 'which {0} || true'.format(name)


class Date(FactBase):
    '''
    Returns the current datetime on the server.
    '''

    command = 'LANG=C date'
    default = datetime.now

    @staticmethod
    def process(output):
        return parse_date(output[0])


class Mounts(FactBase):
    '''
    Returns a dictionary of mounted filesystems and information.

    .. code:: python

        {
            "/": {
                "device": "/dev/mv2",
                "type": "ext4",
                "options": [
                    "rw",
                    "relatime"
                ]
            },
        }
    '''

    command = 'mount'
    default = dict

    @staticmethod
    def process(output):
        devices = {}

        for line in output:
            is_map = False
            if line.startswith('map '):
                line = line[4:]
                is_map = True

            device, _, path, other_bits = line.split(' ', 3)

            if is_map:
                device = 'map {0}'.format(device)

            if other_bits.startswith('type'):
                _, type_, options = other_bits.split(' ', 2)
                options = options.strip('()').split(',')
            else:
                options = other_bits.strip('()').split(',')
                type_ = options.pop(0)

            devices[path] = {
                'device': device,
                'type': type_,
                'options': [option.strip() for option in options],
            }

        return devices


class KernelModules(FactBase):
    '''
    Returns a dictionary of kernel module name -> info.

    .. code:: python

        {
            'module_name': {
                'size': 0,
                'instances': 0,
                'state': 'Live',
            },
        }
    '''

    command = '! test -f /proc/modules || cat /proc/modules'
    default = dict

    @staticmethod
    def process(output):
        modules = {}

        for line in output:
            name, size, instances, depends, state, _ = line.split(' ', 5)
            instances = int(instances)

            module = {
                'size': size,
                'instances': instances,
                'state': state,
            }

            if depends != '-':
                module['depends'] = [
                    value
                    for value in depends.split(',')
                    if value
                ]

            modules[name] = module

        return modules


class LsbRelease(FactBase):
    '''
    Returns a dictionary of release information using ``lsb_release``.

    .. code:: python

        {
            "id": "Ubuntu",
            "description": "Ubuntu 18.04.2 LTS",
            "release": "18.04",
            "codename": "bionic",
            ...
        }
    '''

    command = 'lsb_release -ca'
    requires_command = 'lsb_release'

    @staticmethod
    def process(output):
        items = {}

        for line in output:
            if ':' not in line:
                continue

            key, value = line.split(':', 1)

            key = key.strip().lower()

            # Turn "distributor id" into "id"
            if ' ' in key:
                key = key.split(' ')[-1]

            value = value.strip()

            items[key] = value

        return items


class Sysctl(FactBase):
    '''
    Returns a dictionary of sysctl settings and values.

    .. code:: python

        {
            "fs.inotify.max_queued_events": 16384,
            "fs.inode-state": [
                44565,
                360,
            ],
        }
    '''

    command = 'sysctl -a'
    default = dict

    @staticmethod
    def process(output):
        sysctls = {}

        for line in output:
            key = values = None

            if '=' in line:
                key, values = line.split('=', 1)
            elif ':' in line:
                key, values = line.split(':', 1)
            else:
                continue  # pragma: no cover

            if key and values:
                key = key.strip()
                values = values.strip()

                if re.match(r'^[a-zA-Z0-9_\.\s]+$', values):
                    values = [
                        try_int(item.strip())
                        for item in values.split()
                    ]

                    if len(values) == 1:
                        values = values[0]

                sysctls[key] = values

        return sysctls


class Groups(FactBase):
    '''
    Returns a list of groups on the system.
    '''

    command = 'cat /etc/group'
    default = list

    @staticmethod
    def process(output):
        groups = []

        for line in output:
            if ':' in line:
                groups.append(line.split(':')[0])

        return groups


class Crontab(FactBase):
    '''
    Returns a dictionary of cron command -> execution time.

    .. code:: python

        {
            '/path/to/command': {
                'minute': '*',
                'hour': '*',
                'month': '*',
                'day_of_month': '*',
                'day_of_week': '*',
            },
        }
    '''

    default = dict

    requires_command = 'crontab'

    @staticmethod
    def command(user=None):
        if user:
            return 'crontab -l -u {0}'.format(user)

        return 'crontab -l'

    @staticmethod
    def process(output):
        crons = {}
        current_comments = []

        for line in output:
            line = line.strip()
            if not line or line.startswith('#'):
                current_comments.append(line)
                continue

            minute, hour, day_of_month, month, day_of_week, command = line.split(' ', 5)
            crons[command] = {
                'minute': try_int(minute),
                'hour': try_int(hour),
                'month': try_int(month),
                'day_of_month': try_int(day_of_month),
                'day_of_week': try_int(day_of_week),
                'comments': current_comments,
            }
            current_comments = []

        return crons


class Users(FactBase):
    '''
    Returns a dictionary of users -> details.

    .. code:: python

        {
            'user_name': {
                'comment': 'Full Name',
                'home': '/home/user_name',
                'shell': '/bin/bash,
                'group': 'main_user_group',
                'groups': [
                    'other',
                    'groups'
                ]
            },
        }
    '''

    command = '''
        for i in `cat /etc/passwd | cut -d: -f1`; do
            ID=`id $i`;
            META=`cat /etc/passwd | grep ^$i: | cut -d: -f5-7`;
            echo "$ID $META";
        done
    '''.strip()

    default = dict

    regex = r'^uid=[0-9]+\(([a-zA-Z0-9_\.\-]+)\) gid=[0-9]+\(([a-zA-Z0-9_\.\-]+)\) groups=([a-zA-Z0-9_\.\-,\(\)\s]+) (.*)$'  # noqa
    group_regex = r'^[0-9]+\(([a-zA-Z0-9_\.\-]+)\)$'

    def process(self, output):
        users = {}
        for line in output:
            matches = re.match(self.regex, line)

            if matches:
                # Parse out the comment/home/shell
                comment_home_shell = matches.group(4).split(':')
                comment = comment_home_shell[0] or None
                home = comment_home_shell[1] or None
                shell = comment_home_shell[2] or None

                # Main user group
                group = matches.group(2)

                # Parse the groups
                groups = []
                for group_matches in matches.group(3).split(','):
                    name = re.match(self.group_regex, group_matches.strip())
                    if name:
                        name = name.group(1)
                    else:
                        continue  # pragma: no cover

                    # We only want secondary groups here
                    if name != group:
                        groups.append(name)

                users[matches.group(1)] = {
                    'group': group,
                    'groups': groups,
                    'comment': comment,
                    'home': home,
                    'shell': shell,
                }

        return users


class LinuxDistribution(FactBase):
    '''
    Returns a dict of the Linux distribution version. Ubuntu, Debian, CentOS,
    Fedora & Gentoo currently. Also contains any key/value items located in
    release files.

    .. code:: python

        {
            'name': 'Ubuntu',
            'major': 20,
            'minor': 04,
            'release_meta': {
                'CODENAME': 'focal',
                'ID_LIKE': 'debian',
                ...
            }
        }
    '''
    command = ('cd /etc/ && for file in $(ls -pdL *-release | grep -v /); '
               'do echo "/etc/${file}"; cat "/etc/${file}"; echo ---; '
               'done')

    name_to_pretty_name = {
        'alpine': 'Alpine',
        'centos': 'CentOS',
        'fedora': 'Fedora',
        'gentoo': 'Gentoo',
        'opensuse': 'openSUSE',
        'rhel': 'RedHat',
        'ubuntu': 'Ubuntu',
    }

    @staticmethod
    def default():
        return {
            'name': None,
            'major': None,
            'minor': None,
            'release_meta': {},
        }

    def process(self, output):
        parts = {}
        for part in '\n'.join(output).strip().split('---'):
            if not part.strip():
                continue
            filename, content = part.strip().split('\n', 1)
            parts[filename] = content

        release_info = self.default()
        if not parts:
            return release_info

        temp_root = mkdtemp()
        try:
            temp_etc_dir = os.path.join(temp_root, 'etc')
            os.mkdir(temp_etc_dir)

            for filename, content in parts.items():
                with open(os.path.join(temp_etc_dir, os.path.basename(filename)), 'w') as fp:
                    fp.write(content)

            parsed = get_distro_info(temp_root)

            release_meta = {key.upper(): value for key, value in parsed.os_release_info().items()}
            release_info.update({
                'name': self.name_to_pretty_name.get(parsed.id(), parsed.name()),
                'major': try_int(parsed.major_version()),
                'minor': try_int(parsed.minor_version()) or None,
                'release_meta': release_meta,
            })

        finally:
            shutil.rmtree(temp_root)

        return release_info


class LinuxName(ShortFactBase):
    '''
    Returns the name of the Linux distribution. Shortcut for
    ``host.fact.linux_distribution['name']``.
    '''

    fact = LinuxDistribution

    @staticmethod
    def process_data(data):
        return data['name']


class Selinux(FactBase):
    '''
    Discovers the SELinux related facts on the target host.

    .. code:: python

        {
            'mode': 'enabled',
        }
    '''

    command = 'sestatus'
    requires_command = 'sestatus'

    @staticmethod
    def default():
        return {
            'mode': None,
        }

    def process(self, output):
        selinux_info = self.default()

        match = re.match(r'^SELinux status:\s+(\S+)', '\n'.join(output))

        if not match:
            return selinux_info

        selinux_info['mode'] = match.group(1)

        return selinux_info


class LinuxGui(FactBase):
    '''
    Returns a list of available Linux GUIs.
    '''

    command = 'ls /usr/bin/*session || true'
    default = list

    known_gui_binaries = {
        '/usr/bin/gnome-session': 'GNOME',
        '/usr/bin/mate-session': 'MATE',
        '/usr/bin/lxsession': 'LXDE',
        '/usr/bin/plasma_session': 'KDE Plasma',
        '/usr/bin/xfce4-session': 'XFCE 4',
    }

    def process(self, output):
        gui_names = []

        for line in output:
            gui_name = self.known_gui_binaries.get(line)
            if gui_name:
                gui_names.append(gui_name)

        return gui_names


class HasGui(ShortFactBase):
    '''
    Returns a boolean indicating the remote side has GUI capabilities. Linux only.
    '''

    fact = LinuxGui

    @staticmethod
    def process_data(data):
        return len(data) > 0



'''
The files module handles filesystem state, file uploads and template generation.
'''

from __future__ import unicode_literals

import posixpath
import sys
import traceback

from datetime import timedelta
from fnmatch import fnmatch
from os import makedirs, path as os_path, walk

import six

from jinja2 import TemplateRuntimeError, TemplateSyntaxError, UndefinedError

from pyinfra import logger
from pyinfra.api import (
    FileDownloadCommand,
    FileUploadCommand,
    operation,
    OperationError,
    OperationTypeError,
    RsyncCommand,
)
from pyinfra.api.connectors.util import escape_unix_path
from pyinfra.api.util import get_file_sha1, get_template, memoize

from .util.compat import fspath
from .util.files import (
    chmod,
    chown,
    ensure_mode_int,
    ensure_whole_line_match,
    get_timestamp,
    sed_replace,
)


@operation(pipeline_facts={
    'file': 'dest',
})
def download(
    src, dest,
    user=None, group=None, mode=None, cache_time=None, force=False,
    sha256sum=None, sha1sum=None, md5sum=None,
    state=None, host=None,
):
    '''
    Download files from remote locations using curl or wget.

    + src: source URL of the file
    + dest: where to save the file
    + user: user to own the files
    + group: group to own the files
    + mode: permissions of the files
    + cache_time: if the file exists already, re-download after this time (in seconds)
    + force: always download the file, even if it already exists
    + sha256sum: sha256 hash to checksum the downloaded file against
    + sha1sum: sha1 hash to checksum the downloaded file against
    + md5sum: md5 hash to checksum the downloaded file against

    Example:

    .. code:: python

        files.download(
            name='Download the Docker repo file',
            src='https://download.docker.com/linux/centos/docker-ce.repo',
            dest='/etc/yum.repos.d/docker-ce.repo',
        )
    '''

    dest = escape_unix_path(dest)

    info = host.fact.file(dest)
    # Destination is a directory?
    if info is False:
        raise OperationError(
            'Destination {0} already exists and is not a file'.format(dest),
        )

    # Do we download the file? Force by default
    download = force

    # Doesn't exist, lets download it
    if info is None:
        download = True

    # Destination file exists & cache_time: check when the file was last modified,
    # download if old
    else:
        if cache_time:
            # Time on files is not tz-aware, and will be the same tz as the server's time,
            # so we can safely remove the tzinfo from host.fact.date before comparison.
            cache_time = host.fact.date.replace(tzinfo=None) - timedelta(seconds=cache_time)
            if info['mtime'] and info['mtime'] > cache_time:
                download = True

        if sha1sum:
            if sha1sum != host.fact.sha1_file(dest):
                download = True

        if sha256sum:
            if sha256sum != host.fact.sha256_file(dest):
                download = True

        if md5sum:
            if md5sum != host.fact.md5_file(dest):
                download = True

    # If we download, always do user/group/mode as SSH user may be different
    if download:
        curl_command = 'curl -sSLf {0} -o {1}'.format(src, dest)
        wget_command = 'wget -q {0} -O {1} || (rm -f {1}; exit 1)'.format(src, dest)

        if host.fact.which('curl'):
            yield curl_command
        elif host.fact.which('wget'):
            yield wget_command
        else:
            yield '({0}) || ({1})'.format(curl_command, wget_command)

        if user or group:
            yield chown(dest, user, group)

        if mode:
            yield chmod(dest, mode)

        if sha1sum:
            yield (
                '((sha1sum {0} 2> /dev/null || shasum {0} || sha1 {0}) | grep {1}) '
                '|| (echo "SHA1 did not match!" && exit 1)'
            ).format(dest, sha1sum)

        if sha256sum:
            yield (
                '((sha256sum {0} 2> /dev/null || shasum -a 256 {0} || sha256 {0}) | grep {1}) '
                '|| (echo "SHA256 did not match!" && exit 1)'
            ).format(dest, sha256sum)

        if md5sum:
            yield (
                '((md5sum {0} 2> /dev/null || md5 {0}) | grep {1}) '
                '|| (echo "MD5 did not match!" && exit 1)'
            ).format(dest, md5sum)

    else:
        host.noop('file {0} has already been downloaded'.format(dest))


@operation
def line(
    path, line,
    present=True, replace=None, flags=None, backup=False,
    interpolate_variables=False,
    state=None, host=None,
    assume_present=False,
):
    '''
    Ensure lines in files using grep to locate and sed to replace.

    + path: target remote file to edit
    + line: string or regex matching the target line
    + present: whether the line should be in the file
    + replace: text to replace entire matching lines when ``present=True``
    + flags: list of flags to pass to sed when replacing/deleting
    + backup: whether to backup the file (see below)
    + interpolate_variables: whether to interpolate variables in ``replace``

    Regex line matching:
        Unless line matches a line (starts with ^, ends $), pyinfra will wrap it such that
        it does, like: ``^.*LINE.*$``. This means we don't swap parts of lines out. To
        change bits of lines, see ``files.replace``.

    Regex line escaping:
        If matching special characters (eg a crontab line containing *), remember to escape
        it first using Python's ``re.escape``.

    Backup:
        If set to ``True``, any editing of the file will place an old copy with the ISO
        date (taken from the machine running ``pyinfra``) appended as the extension. If
        you pass a string value this will be used as the extension of the backed up file.

    Append:
        If ``line`` is not in the file but we want it (``present`` set to ``True``), then
        it will be append to the end of the file.

    Examples:

    .. code:: python

        # prepare to do some maintenance
        maintenance_line = 'SYSTEM IS DOWN FOR MAINTENANCE'
        files.line(
            name='Add the down-for-maintence line in /etc/motd',
            path='/etc/motd',
            line=maintenance_line,
        )

        # Then, after the mantenance is done, remove the maintenance line
        files.line(
            name='Remove the down-for-maintenance line in /etc/motd',
            path='/etc/motd',
            line=maintenance_line,
            replace='',
            present=False,
        )

        # example where there is '*' in the line
        files.line(
            name='Ensure /netboot/nfs is in /etc/exports',
            path='/etc/exports',
            line=r'/netboot/nfs .*',
            replace='/netboot/nfs *(ro,sync,no_wdelay,insecure_locks,no_root_squash,'
            'insecure,no_subtree_check)',
        )

        files.line(
            name='Ensure myweb can run /usr/bin/python3 without password',
            path='/etc/sudoers',
            line=r'myweb .*',
            replace='myweb ALL=(ALL) NOPASSWD: /usr/bin/python3',
        )

        # example when there are double quotes (")
        line = 'QUOTAUSER=""'
        files.line(
            name='Example with double quotes (")',
            path='/etc/adduser.conf',
            line='^{}$'.format(line),
            replace=line,
        )
    '''

    path = escape_unix_path(path)

    match_line = ensure_whole_line_match(line)

    # Is there a matching line in this file?
    if assume_present:
        present_lines = [line]
    else:
        present_lines = host.fact.find_in_file(path, match_line)

    # If replace present, use that over the matching line
    if replace:
        line = replace
    # We must provide some kind of replace to sed_replace_command below
    else:
        replace = ''

    # Save commands for re-use in dynamic script when file not present at fact stage
    make_backup_command = ''
    if backup:
        make_backup_command = 'cp {0} {0}.{1} && '.format(path, get_timestamp())

    echo_command = (
        '{0}echo "{1}" >> {2}'.format(make_backup_command, line, path)
        if interpolate_variables else
        "{0}echo '{1}' >> {2}".format(make_backup_command, line, path)
    )
    sed_replace_command = sed_replace(
        path, match_line, replace,
        flags=flags,
        backup=backup,
        interpolate_variables=interpolate_variables,
    )

    # No line and we want it, append it
    if not present_lines and present:
        quoted_match_line = (
            '"{0}"'.format(match_line)
            if interpolate_variables else
            "'{0}'".format(match_line)
        )

        # If the file does not exist - it *might* be created, so we handle it
        # dynamically with a little script.
        if present_lines is None:
            yield '''
                if [ -f '{target}' ]; then
                    (grep {quoted_match_line} '{target}' && {sed_replace_command}) 2> /dev/null || \
                    {echo_command};
                else
                    {echo_command};
                fi
            '''.format(
                target=path,
                quoted_match_line=quoted_match_line,
                echo_command=echo_command,
                sed_replace_command=sed_replace_command,
            )

        # File exists but has no matching lines - append it.
        else:
            # If we're doing replacement, only append if the *replacement* line
            # does not exist (as we are appending the replacement).
            if replace:
                present_lines = host.fact.find_in_file(
                    path, ensure_whole_line_match(replace),
                )

            if not present_lines:
                yield echo_command
            else:
                host.noop('line "{0}" exists in {1}'.format(replace or line, path))

    # Line(s) exists and we want to remove them, replace with nothing
    elif present_lines and not present:
        yield sed_replace(
            path, match_line, '',
            flags=flags,
            backup=backup,
            interpolate_variables=interpolate_variables,
        )

    # Line(s) exists and we have want to ensure they're correct
    elif present_lines and present:
        # If any of lines are different, sed replace them
        if replace and any(line != replace for line in present_lines):
            yield sed_replace_command
        else:
            host.noop('line "{0}" exists in {1}'.format(replace or line, path))


@operation
def replace(
    path, match, replace,
    flags=None, backup=False,
    interpolate_variables=False,
    state=None, host=None,
):
    '''
    Replace contents of a file using ``sed``.

    + path: target remote file to edit
    + match: text/regex to match for
    + replace: text to replace with
    + flags: list of flags to pass to sed
    + backup: whether to backup the file (see below)
    + interpolate_variables: whether to interpolate variables in ``replace``

    Backup:
        If set to ``True``, any editing of the file will place an old copy with the ISO
        date (taken from the machine running ``pyinfra``) appended as the extension. If
        you pass a string value this will be used as the extension of the backed up file.

    Example:

    .. code:: python

        files.replace(
            name='Change part of a line in a file',
            path='/etc/motd',
            match='verboten',
            replace='forbidden',
        )
    '''

    path = escape_unix_path(path)

    existing_lines = host.fact.find_in_file(path, match)

    # Only do the replacement if the file does not exist (it may be created earlier)
    # or we have matching lines.
    if existing_lines is None or existing_lines:
        yield sed_replace(
            path, match, replace,
            flags=flags,
            backup=backup,
            interpolate_variables=interpolate_variables,
        )
    else:
        host.noop('string "{0}" does not exist in {1}'.format(match, path))


@operation(pipeline_facts={
    'find_files': 'destination',
})
def sync(
    src, dest,
    user=None, group=None, mode=None, delete=False,
    exclude=None, exclude_dir=None, add_deploy_dir=True,
    state=None, host=None,
):
    '''
    Syncs a local directory with a remote one, with delete support. Note that delete will
    remove extra files on the remote side, but not extra directories.

    + src: local directory to sync
    + dest: remote directory to sync to
    + user: user to own the files and directories
    + group: group to own the files and directories
    + mode: permissions of the files
    + delete: delete remote files not present locally
    + exclude: string or list/tuple of strings to match & exclude files (eg *.pyc)
    + exclude_dir: string or list/tuple of strings to match & exclude directories (eg node_modules)

    Example:

    .. code:: python

        # Sync local files/tempdir to remote /tmp/tempdir
        files.sync(
            name='Sync a local directory with remote',
            src='files/tempdir',
            dest='/tmp/tempdir',
        )
    '''

    # If we don't enforce the source ending with /, remote_dirname below might start with
    # a /, which makes the os_path.join cut off the destination bit.
    if not src.endswith(os_path.sep):
        src = '{0}{1}'.format(src, os_path.sep)

    # Add deploy directory?
    if add_deploy_dir and state.deploy_dir:
        src = os_path.join(state.deploy_dir, src)

    # Ensure the source directory exists
    if not os_path.isdir(src):
        raise IOError('No such directory: {0}'.format(src))

    # Ensure exclude is a list/tuple
    if exclude is not None:
        if not isinstance(exclude, (list, tuple)):
            exclude = [exclude]

    # Ensure exclude_dir is a list/tuple
    if exclude_dir is not None:
        if not isinstance(exclude_dir, (list, tuple)):
            exclude_dir = [exclude_dir]

    put_files = []
    ensure_dirnames = []
    for dirname, _, filenames in walk(src):
        remote_dirname = dirname.replace(src, '')

        # Should we exclude this dir?
        if exclude_dir and any(fnmatch(remote_dirname, match) for match in exclude_dir):
            continue

        if remote_dirname:
            ensure_dirnames.append(remote_dirname)

        for filename in filenames:
            full_filename = os_path.join(dirname, filename)

            # Should we exclude this file?
            if exclude and any(fnmatch(full_filename, match) for match in exclude):
                continue

            put_files.append((
                # Join local as normal (unix, win)
                full_filename,
                # Join remote as unix like
                '/'.join(
                    item for item in
                    (dest, remote_dirname, filename)
                    if item
                ),
            ))

    # Ensure the destination directory - if the destination is a link, ensure
    # the link target is a directory.
    dest_to_ensure = dest
    dest_link_info = host.fact.link(dest)
    if dest_link_info:
        dest_to_ensure = dest_link_info['link_target']

    yield directory(
        dest_to_ensure,
        user=user, group=group,
        state=state, host=host,
    )

    # Ensure any remote dirnames
    for dirname in ensure_dirnames:
        yield directory(
            '/'.join((dest, dirname)),
            user=user, group=group,
            state=state, host=host,
        )

    # Put each file combination
    for local_filename, remote_filename in put_files:
        yield put(
            local_filename, remote_filename,
            user=user, group=group, mode=mode,
            add_deploy_dir=False,
            create_remote_dir=False,  # handled above
            state=state, host=host,
        )

    # Delete any extra files
    if delete:
        remote_filenames = set(host.fact.find_files(dest) or [])
        wanted_filenames = set([remote_filename for _, remote_filename in put_files])
        files_to_delete = remote_filenames - wanted_filenames
        for filename in files_to_delete:
            # Should we exclude this file?
            if exclude and any(fnmatch(filename, match) for match in exclude):
                continue

            yield file(filename, present=False, state=state, host=host)


@memoize
def show_rsync_warning():
    logger.warning('The `files.rsync` operation is in alpha!')


@operation
def rsync(
    src, dest, flags=['-ax', '--delete'],
    state=None, host=None,
):
    '''
    Use ``rsync`` to sync a local directory to the remote system. This operation will actually call
    the ``rsync`` binary on your system.

    .. important::
        The ``files.rsync`` operation is in alpha, and only supported using SSH
        or ``@local`` connectors.

    .. caution::
        When using SSH, the ``files.rsync`` operation only supports the ``sudo`` and ``sudo_user``
        global arguments.
    '''

    show_rsync_warning()

    try:
        host.check_can_rsync()
    except NotImplementedError as e:
        raise OperationError(*e.args)

    yield RsyncCommand(src, dest, flags)


def _create_remote_dir(state, host, remote_filename, user, group):
    # Always use POSIX style path as local might be Windows, remote always *nix
    remote_dirname = posixpath.dirname(remote_filename)
    if remote_dirname:
        yield directory(
            path=remote_dirname,
            user=user, group=group,
            _no_check_owner_mode=True,  # don't check existing user/mode
            _no_fail_on_link=True,  # don't fail if the path is a link
            state=state, host=host,
        )


@operation(pipeline_facts={
    'file': 'src',
    'sha1_file': 'src',
})
def get(
    src, dest,
    add_deploy_dir=True, create_local_dir=False, force=False,
    state=None, host=None,
):
    '''
    Download a file from the remote system.

    + src: the remote filename to download
    + dest: the local filename to download the file to
    + add_deploy_dir: dest is relative to the deploy directory
    + create_local_dir: create the local directory if it doesn't exist
    + force: always download the file, even if the local copy matches

    Note:
        This operation is not suitable for large files as it may involve copying
        the remote file before downloading it.

    Example:

    .. code:: python

        files.get(
            name='Download a file from a remote',
            src='/etc/centos-release',
            dest='/tmp/whocares',
        )
    '''

    src = escape_unix_path(src)

    if add_deploy_dir and state.deploy_dir:
        dest = os_path.join(state.deploy_dir, dest)

    if create_local_dir:
        local_pathname = os_path.dirname(dest)
        if not os_path.exists(local_pathname):
            makedirs(local_pathname)

    remote_file = host.fact.file(src)

    # No remote file, so assume exists and download it "blind"
    if not remote_file or force:
        yield FileDownloadCommand(src, dest)

    # No local file, so always download
    elif not os_path.exists(dest):
        yield FileDownloadCommand(src, dest)

    # Remote file exists - check if it matches our local
    else:
        local_sum = get_file_sha1(dest)
        remote_sum = host.fact.sha1_file(src)

        # Check sha1sum, upload if needed
        if local_sum != remote_sum:
            yield FileDownloadCommand(src, dest)


@operation(pipeline_facts={
    'file': 'dest',
    'sha1_file': 'dest',
})
def put(
    src, dest,
    user=None, group=None, mode=None, add_deploy_dir=True,
    create_remote_dir=True, force=False, assume_exists=False,
    state=None, host=None,
):
    '''
    Upload a local file to the remote system.

    + src: local filename to upload
    + dest: remote filename to upload to
    + user: user to own the files
    + group: group to own the files
    + mode: permissions of the files
    + add_deploy_dir: src is relative to the deploy directory
    + create_remote_dir: create the remote directory if it doesn't exist
    + force: always upload the file, even if the remote copy matches
    + assume_exists: whether to assume the local file exists

    ``create_remote_dir``:
        If the remote directory does not exist it will be created using the same
        user & group as passed to ``files.put``. The mode will *not* be copied over,
        if this is required call ``files.directory`` separately.

    Note:
        This operation is not suitable for large files as it may involve copying
        the file before uploading it.

    Examples:

    .. code:: python

        # Note: This requires a 'files/motd' file on the local filesystem
        files.put(
            name='Update the message of the day file',
            src='files/motd',
            dest='/etc/motd',
            mode='644',
        )
    '''

    dest = escape_unix_path(dest)

    # Upload IO objects as-is
    if hasattr(src, 'read'):
        local_file = src

    # Assume string filename
    else:
        # Add deploy directory?
        if add_deploy_dir and state.deploy_dir:
            src = os_path.join(state.deploy_dir, src)

        local_file = src

        if not assume_exists and not os_path.isfile(local_file):
            raise IOError('No such file: {0}'.format(local_file))

    mode = ensure_mode_int(mode)
    remote_file = host.fact.file(dest)

    if create_remote_dir:
        yield _create_remote_dir(state, host, dest, user, group)

    # No remote file, always upload and user/group/mode if supplied
    if not remote_file or force:
        yield FileUploadCommand(local_file, dest)

        if user or group:
            yield chown(dest, user, group)

        if mode:
            yield chmod(dest, mode)

    # File exists, check sum and check user/group/mode if supplied
    else:
        local_sum = get_file_sha1(src)
        remote_sum = host.fact.sha1_file(dest)

        # Check sha1sum, upload if needed
        if local_sum != remote_sum:
            yield FileUploadCommand(local_file, dest)

            if user or group:
                yield chown(dest, user, group)

            if mode:
                yield chmod(dest, mode)

        else:
            changed = False

            # Check mode
            if mode and remote_file['mode'] != mode:
                yield chmod(dest, mode)
                changed = True

            # Check user/group
            if (
                (user and remote_file['user'] != user)
                or (group and remote_file['group'] != group)
            ):
                yield chown(dest, user, group)
                changed = True

            if not changed:
                host.noop('file {0} is already uploaded'.format(dest))


@operation
def template(
    src, dest,
    user=None, group=None, mode=None, create_remote_dir=True,
    state=None, host=None,
    **data
):
    '''
    Generate a template using jinja2 and write it to the remote system.

    + src: local template filename
    + dest: remote filename
    + user: user to own the files
    + group: group to own the files
    + mode: permissions of the files
    + create_remote_dir: create the remote directory if it doesn't exist

    ``create_remote_dir``:
        If the remote directory does not exist it will be created using the same
        user & group as passed to ``files.put``. The mode will *not* be copied over,
        if this is required call ``files.directory`` separately.

    Notes:
       Common convention is to store templates in a "templates" directory and
       have a filename suffix with '.j2' (for jinja2).

       For information on the template syntax, see
       `the jinja2 docs <https://jinja.palletsprojects.com>`_.

    Examples:

    .. code:: python

        files.template(
            name='Create a templated file',
            src='templates/somefile.conf.j2',
            dest='/etc/somefile.conf',
        )

        files.template(
            name='Create service file',
            src='templates/myweb.service.j2',
            dest='/etc/systemd/system/myweb.service',
            mode='755',
            user='root',
            group='root',
        )

        # Example showing how to pass python variable to template file.
        # The .j2 file can use `{{ foo_variable }}` to be interpolated.
        foo_variable = 'This is some foo variable contents'
        files.template(
            name='Create a templated file',
            src='templates/foo.j2',
            dest='/tmp/foo',
            foo_variable=foo_variable,
        )
    '''

    dest = escape_unix_path(dest)

    if state.deploy_dir:
        src = os_path.join(state.deploy_dir, src)

    # Ensure host/state/inventory are available inside templates (if not set)
    data.setdefault('host', host)
    data.setdefault('state', state)
    data.setdefault('inventory', state.inventory)

    # Render and make file-like it's output
    try:
        output = get_template(src).render(data)
    except (TemplateRuntimeError, TemplateSyntaxError, UndefinedError) as e:
        trace_frames = traceback.extract_tb(sys.exc_info()[2])
        trace_frames = [
            frame for frame in trace_frames
            if frame[2] in ('template', '<module>', 'top-level template code')
        ]  # thank you https://github.com/saltstack/salt/blob/master/salt/utils/templates.py

        line_number = trace_frames[-1][1]

        # Quickly read the line in question and one above/below for nicer debugging
        with open(src, 'r') as f:
            template_lines = f.readlines()

        template_lines = [line.strip() for line in template_lines]
        relevant_lines = template_lines[max(line_number - 2, 0):line_number + 1]

        raise OperationError('Error in template: {0} (L{1}): {2}\n...\n{3}\n...'.format(
            src, line_number, e, '\n'.join(relevant_lines),
        ))

    output_file = six.StringIO(output)
    # Set the template attribute for nicer debugging
    output_file.template = src

    # Pass to the put function
    yield put(
        output_file, dest,
        user=user, group=group, mode=mode,
        add_deploy_dir=False,
        create_remote_dir=create_remote_dir,
        state=state, host=host,
    )


def _validate_path(path):
    try:
        path = fspath(path)
    except TypeError:
        raise OperationTypeError('`path` must be a string or `os.PathLike` object')


@operation(pipeline_facts={
    'link': 'path',
})
def link(
    path,
    target=None, present=True, assume_present=False,
    user=None, group=None, symbolic=True,
    create_remote_dir=True,
    state=None, host=None,
):
    '''
    Add/remove/update links.

    + path: the name of the link
    + target: the file/directory the link points to
    + present: whether the link should exist
    + assume_present: whether to assume the link exists
    + user: user to own the link
    + group: group to own the link
    + symbolic: whether to make a symbolic link (vs hard link)
    + create_remote_dir: create the remote directory if it doesn't exist

    ``create_remote_dir``:
        If the remote directory does not exist it will be created using the same
        user & group as passed to ``files.put``. The mode will *not* be copied over,
        if this is required call ``files.directory`` separately.

    Source changes:
        If the link exists and points to a different target, pyinfra will remove it and
        recreate a new one pointing to then new target.

    Examples:

    .. code:: python

        # simple example showing how to link to a file
        files.link(
            name='Create link /etc/issue2 that points to /etc/issue',
            path='/etc/issue2',
            target='/etc/issue',
        )


        # complex example demonstrating the assume_present option
        from pyinfra.operations import apt, files

        install_nginx = apt.packages(
            name='Install nginx',
            packages=['nginx'],
        )

        files.link(
            name='Remove default nginx site',
            path='/etc/nginx/sites-enabled/default',
            present=False,
            assume_present=install_nginx.changed,
        )
    '''

    _validate_path(path)

    if present and not target:
        raise OperationError('If present is True target must be provided')

    path = escape_unix_path(path)
    info = host.fact.link(path)

    # Not a link?
    if info is False:
        raise OperationError('{0} exists and is not a link'.format(path))

    add_cmd = 'ln{0} {1} {2}'.format(
        ' -s' if symbolic else '',
        target, path,
    )

    remove_cmd = 'rm -f {0}'.format(path)

    # No link and we want it
    if not assume_present and info is None and present:
        if create_remote_dir:
            yield _create_remote_dir(state, host, path, user, group)

        yield add_cmd

        if user or group:
            yield chown(path, user, group, dereference=False)

        host.fact._create(
            'link',
            args=(path,),
            data={'link_target': target, 'group': group, 'user': user},
        )

    # It exists and we don't want it
    elif (assume_present or info) and not present:
        yield remove_cmd
        host.fact._delete('link', args=(path,))

    # Exists and want to ensure it's state
    elif (assume_present or info) and present:
        if assume_present and not info:
            info = {'link_target': None, 'group': None, 'user': None}
            host.fact._create('link', args=(path,), data=info)

        # If we have an absolute path - prepend to any non-absolute values from the fact
        # and/or the source.
        if os_path.isabs(path):
            link_dirname = os_path.dirname(path)

            if not os_path.isabs(target):
                target = os_path.normpath('/'.join((link_dirname, target)))

            if info and not os_path.isabs(info['link_target']):
                info['link_target'] = os_path.normpath(
                    '/'.join((link_dirname, info['link_target'])),
                )

        changed = False

        # If the target is wrong, remove & recreate the link
        if not info or info['link_target'] != target:
            changed = True
            yield remove_cmd
            yield add_cmd
            info['link_target'] = target

        # Check user/group
        if (
            (not info and (user or group))
            or (user and info['user'] != user)
            or (group and info['group'] != group)
        ):
            yield chown(path, user, group, dereference=False)
            changed = True
            if user:
                info['user'] = user
            if group:
                info['group'] = group

        if not changed:
            host.noop('link {0} already exists'.format(path))


@operation(pipeline_facts={
    'file': 'path',
})
def file(
    path,
    present=True, assume_present=False,
    user=None, group=None, mode=None, touch=False,
    create_remote_dir=True,
    state=None, host=None,
):
    '''
    Add/remove/update files.

    + path: name/path of the remote file
    + present: whether the file should exist
    + assume_present: whether to assume the file exists
    + user: user to own the files
    + group: group to own the files
    + mode: permissions of the files as an integer, eg: 755
    + touch: whether to touch the file
    + create_remote_dir: create the remote directory if it doesn't exist

    ``create_remote_dir``:
        If the remote directory does not exist it will be created using the same
        user & group as passed to ``files.put``. The mode will *not* be copied over,
        if this is required call ``files.directory`` separately.

    Example:

    .. code:: python

        # Note: The directory /tmp/secret will get created with the default umask.
        files.file(
            name='Create /tmp/secret/file',
            path='/tmp/secret/file',
            mode='600',
            user='root',
            group='root',
            touch=True,
            create_remote_dir=True,
        )
    '''

    _validate_path(path)

    mode = ensure_mode_int(mode)
    path = escape_unix_path(path)
    info = host.fact.file(path)

    # Not a file?!
    if info is False:
        raise OperationError('{0} exists and is not a file'.format(path))

    # Doesn't exist & we want it
    if not assume_present and info is None and present:
        if create_remote_dir:
            yield _create_remote_dir(state, host, path, user, group)

        yield 'touch {0}'.format(path)

        if mode:
            yield chmod(path, mode)
        if user or group:
            yield chown(path, user, group)

        host.fact._create(
            'file',
            args=(path,),
            data={'mode': mode, 'group': group, 'user': user},
        )

    # It exists and we don't want it
    elif (assume_present or info) and not present:
        yield 'rm -f {0}'.format(path)
        host.fact._delete('file', args=(path,))

    # It exists & we want to ensure its state
    elif (assume_present or info) and present:
        if assume_present and not info:
            info = {'mode': None, 'group': None, 'user': None}
            host.fact._create('file', args=(path,), data=info)

        changed = False

        if touch:
            changed = True
            yield 'touch {0}'.format(path)

        # Check mode
        if mode and (not info or info['mode'] != mode):
            yield chmod(path, mode)
            info['mode'] = mode
            changed = True

        # Check user/group
        if (
            (not info and (user or group))
            or (user and info['user'] != user)
            or (group and info['group'] != group)
        ):
            yield chown(path, user, group)
            changed = True
            if user:
                info['user'] = user
            if group:
                info['group'] = group

        if not changed:
            host.noop('file {0} already exists'.format(path))


@operation(pipeline_facts={
    'directory': 'path',
})
def directory(
    path,
    present=True, assume_present=False,
    user=None, group=None, mode=None, recursive=False,
    _no_check_owner_mode=False,
    _no_fail_on_link=False,
    state=None, host=None,
):
    '''
    Add/remove/update directories.

    + path: path of the remote folder
    + present: whether the folder should exist
    + assume_present: whether to assume the directory exists
    + user: user to own the folder
    + group: group to own the folder
    + mode: permissions of the folder
    + recursive: recursively apply user/group/mode

    Examples:

    .. code:: python

        files.directory(
            name='Ensure the /tmp/dir_that_we_want_removed is removed',
            path='/tmp/dir_that_we_want_removed',
            present=False,
        )

        files.directory(
            name='Ensure /web exists',
            path='/web',
            user='myweb',
            group='myweb',
        )

        # multiple directories
        dirs = ['/netboot/tftp', '/netboot/nfs']
        for dir in dirs:
            files.directory(
                name='Ensure the directory `{}` exists'.format(dir),
                path=dir,
            )
    '''

    _validate_path(path)

    mode = ensure_mode_int(mode)
    path = escape_unix_path(path)
    info = host.fact.directory(path)

    # Not a directory?!
    if info is False:
        if _no_fail_on_link and host.fact.link(path):
            host.noop('directory {0} already exists (as a link)'.format(path))
            return
        raise OperationError('{0} exists and is not a directory'.format(path))

    # Doesn't exist & we want it
    if not assume_present and info is None and present:
        yield 'mkdir -p {0}'.format(path)
        if mode:
            yield chmod(path, mode, recursive=recursive)
        if user or group:
            yield chown(path, user, group, recursive=recursive)

        host.fact._create(
            'directory',
            args=(path,),
            data={'mode': mode, 'group': group, 'user': user},
        )

    # It exists and we don't want it
    elif (assume_present or info) and not present:
        yield 'rm -rf {0}'.format(path)
        host.fact._delete('directory', args=(path,))

    # It exists & we want to ensure its state
    elif (assume_present or info) and present:
        if assume_present and not info:
            info = {'mode': None, 'group': None, 'user': None}
            host.fact._create('directory', args=(path,), data=info)

        if _no_check_owner_mode:
            return

        changed = False

        if mode and (not info or info['mode'] != mode):
            yield chmod(path, mode, recursive=recursive)
            info['mode'] = mode
            changed = True

        if (
            (not info and (user or group))
            or (user and info['user'] != user)
            or (group and info['group'] != group)
        ):
            yield chown(path, user, group, recursive=recursive)
            changed = True
            if user:
                info['user'] = user
            if group:
                info['group'] = group

        if not changed:
            host.noop('directory {0} already exists'.format(path))



from glob import glob
from os import path


module_filenames = glob(path.join(path.dirname(__file__), '*.py'))
module_names = [path.basename(name)[:-3] for name in module_filenames]
__all__ = [name for name in module_names if name != '__init__']


# Actually triggers __all__ imports to builds facts index
from . import * # noqa



from unittest import TestCase

from click.testing import CliRunner

from pyinfra_cli.exceptions import CliError
from pyinfra_cli.main import cli


class TestCliExceptions(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.test_cli = CliRunner()
        cls.old_cli_show = CliError.show

    @classmethod
    def tearDownClass(cls):
        CliError.show = cls.old_cli_show

    def setUp(self):
        self.exception = None
        CliError.show = lambda e: self.capture_cli_error(e)

    def capture_cli_error(self, e):
        self.exception = e
        self.old_cli_show()

    def assert_cli_exception(self, args, message):
        self.test_cli.invoke(cli, args)

        self.assertIsInstance(self.exception, CliError)
        assert self.exception.message == message

    def test_bad_deploy_file(self):
        self.assert_cli_exception(
            ['my-server.net', 'nop.py'],
            'No deploy file: nop.py',
        )

    def test_bad_fact(self):
        self.assert_cli_exception(
            ['my-server.net', 'fact', 'thing'],
            'No fact: thing',
        )



from __future__ import print_function, unicode_literals

from os import path

import click
import six

import pyinfra

from . import logger, pseudo_state
from .api.connectors.util import run_local_process, split_combined_output
from .api.exceptions import PyinfraError


def include(filename):
    '''
    Executes a local python file within the ``pyinfra.pseudo_state.deploy_dir``
    directory.
    '''

    if not pyinfra.is_cli:
        raise PyinfraError('local.include is only available in CLI mode.')

    if pseudo_state.deploy_dir:
        filename = path.join(pseudo_state.deploy_dir, filename)

    logger.debug('Including local file: {0}'.format(filename))

    try:
        # Fixes a circular import because `pyinfra.local` is really a CLI
        # only thing (so should be `pyinfra_cli.local`). It is kept here
        # to maintain backwards compatibility and the nicer public import
        # (ideally users never need to import from `pyinfra_cli`).

        from pyinfra_cli.config import extract_file_config
        from pyinfra_cli.util import exec_file

        # Load any config defined in the file and setup like a @deploy
        config_data = extract_file_config(filename)
        kwargs = {
            key.lower(): value
            for key, value in six.iteritems(config_data)
            if key in [
                'SUDO', 'SUDO_USER', 'SU_USER',
                'PRESERVE_SUDO_ENV', 'IGNORE_ERRORS',
            ]
        }
        with pseudo_state.deploy(filename, kwargs, None, in_deploy=False):
            exec_file(filename)

        # One potential solution to the above is to add local as an actual
        # module, ie `pyinfra.operations.local`.

    except Exception as e:
        raise PyinfraError(
            'Could not include local file: {0}:\n{1}'.format(filename, e),
        )


def shell(commands, splitlines=False, ignore_errors=False, print_output=False, print_input=False):
    '''
    Subprocess based implementation of pyinfra/api/ssh.py's ``run_shell_command``.

    Args:
        commands (string, list): command or list of commands to execute
        spltlines (bool): optionally have the output split by lines
        ignore_errors (bool): ignore errors when executing these commands
    '''

    if isinstance(commands, six.string_types):
        commands = [commands]

    all_stdout = []

    # Checking for pseudo_state means this function works outside a deploy
    # e.g.: the vagrant connector.
    if pseudo_state.isset():
        print_output = pseudo_state.print_output
        print_input = pseudo_state.print_input

    for command in commands:
        print_prefix = 'localhost: '

        if print_input:
            click.echo('{0}>>> {1}'.format(print_prefix, command), err=True)

        return_code, combined_output = run_local_process(
            command,
            print_output=print_output,
            print_prefix=print_prefix,
        )
        stdout, stderr = split_combined_output(combined_output)

        if return_code > 0 and not ignore_errors:
            raise PyinfraError(
                'Local command failed: {0}\n{1}'.format(command, stderr),
            )

        all_stdout.extend(stdout)

    if not splitlines:
        return '\n'.join(all_stdout)

    return all_stdout



import os

from tempfile import mkstemp

import click
import six

from pyinfra import logger
from pyinfra.api import QuoteString, StringCommand
from pyinfra.api.exceptions import ConnectError, InventoryError, PyinfraError
from pyinfra.api.util import get_file_io, memoize
from pyinfra.progress import progress_spinner

from . import ssh
from .util import make_unix_command


def remote_remove(state, host, filename, print_output=False, print_input=False):
    '''
    Deletes a file on a remote machine over ssh.
    '''
    remove_status, _, remove_stderr = ssh.run_shell_command(
        state, host,
        'rm -f {0}'.format(filename),
        print_output=print_output,
        print_input=print_input)

    if not remove_status:
        raise IOError('\n'.join(remove_stderr))


@memoize
def show_warning():
    logger.warning('The @dockerssh connector is in beta!')


def make_names_data(host_image_str):
    try:
        hostname, image = host_image_str.split(':', 1)
    except (AttributeError, ValueError):  # failure to parse the host_image_str
        raise InventoryError('No ssh host or docker base image provided!')

    if not image:
        raise InventoryError('No docker base image provided!')

    show_warning()

    yield ('@dockerssh/{0}:{1}'.format(hostname, image),
           {'ssh_hostname': hostname, 'docker_image': image},
           ['@dockerssh'])


def connect(state, host):
    if not host.connection:
        host.connection = ssh.connect(state, host)

    if 'docker_container_id' in host.host_data:  # user can provide a docker_container_id
        return host.connection

    try:
        with progress_spinner({'docker run'}):
            # last line is the container ID
            status, stdout, stderr = ssh.run_shell_command(
                state, host,
                'docker run -d {0} tail -f /dev/null'.format(host.data.docker_image),
            )
            if not status:
                raise IOError('\n'.join(stderr))
            container_id = stdout[-1]

    except PyinfraError as e:
        host.connection = None  # fail connection
        raise ConnectError(e.args[0])

    host.host_data['docker_container_id'] = container_id
    return host.connection


def disconnect(state, host):
    container_id = host.host_data['docker_container_id'][:12]

    with progress_spinner({'docker commit'}):
        image_id = ssh.run_shell_command(
            state, host,
            'docker commit {0}'.format(container_id),
        )[1][-1][7:19]  # last line is the image ID, get sha256:[XXXXXXXXXX]...

    with progress_spinner({'docker rm'}):
        ssh.run_shell_command(
            state, host,
            'docker rm -f {0}'.format(container_id),
        )

    logger.info('{0}docker build complete, image ID: {1}'.format(
        host.print_prefix, click.style(image_id, bold=True),
    ))


def run_shell_command(
    state, host, command,
    get_pty=False,
    timeout=None,
    stdin=None,
    success_exit_codes=None,
    print_output=False,
    print_input=False,
    return_combined_output=False,
    **command_kwargs
):
    container_id = host.host_data['docker_container_id']

    # Don't sudo/su in Docker - is this the right thing to do? Makes deploys that
    # target SSH systems work w/Docker out of the box (ie most docker commands
    # are run as root).
    for key in ('sudo', 'su_user'):
        command_kwargs.pop(key, None)

    command = make_unix_command(command, **command_kwargs)
    command = QuoteString(command)

    docker_flags = '-it' if get_pty else '-i'
    docker_command = StringCommand(
        'docker', 'exec', docker_flags, container_id,
        'sh', '-c', command,
    )

    return ssh.run_shell_command(
        state, host,
        docker_command,
        timeout=timeout,
        stdin=stdin,
        success_exit_codes=success_exit_codes,
        print_output=print_output,
        print_input=print_input,
        return_combined_output=return_combined_output,
    )


def put_file(
    state, host, filename_or_io, remote_filename,
    print_output=False, print_input=False,
    **kwargs  # ignored (sudo/etc)
):
    '''
    Upload a file/IO object to the target Docker container by copying it to a
    temporary location and then uploading it into the container using ``docker cp``.
    '''

    fd, temp_filename = mkstemp()
    remote_temp_filename = state.get_temp_filename(temp_filename)

    # Load our file or IO object and write it to the temporary file
    with get_file_io(filename_or_io) as file_io:
        with open(temp_filename, 'wb') as temp_f:
            data = file_io.read()

            if isinstance(data, six.text_type):
                data = data.encode()

            temp_f.write(data)

    # upload file to remote server
    ssh_status = ssh.put_file(state, host, temp_filename, remote_temp_filename)
    if not ssh_status:
        raise IOError('Failed to copy file over ssh')

    try:
        docker_id = host.host_data['docker_container_id']
        docker_command = 'docker cp {0} {1}:{2}'.format(
            remote_temp_filename,
            docker_id,
            remote_filename,
        )

        status, _, stderr = ssh.run_shell_command(
            state, host,
            docker_command,
            print_output=print_output,
            print_input=print_input,
        )
    finally:
        os.close(fd)
        os.remove(temp_filename)
        remote_remove(
            state, host, temp_filename,
            print_output=print_output,
            print_input=print_input,
        )

    if not status:
        raise IOError('\n'.join(stderr))

    if print_output:
        click.echo('{0}file uploaded to container: {1}'.format(
            host.print_prefix, remote_filename,
        ), err=True)

    return status


def get_file(
    state, host, remote_filename, filename_or_io,
    print_output=False, print_input=False,
    **kwargs  # ignored (sudo/etc)
):
    '''
    Download a file from the target Docker container by copying it to a temporary
    location and then reading that into our final file/IO object.
    '''

    temp_filename = state.get_temp_filename(remote_filename)

    try:
        docker_id = host.host_data['docker_container_id']
        docker_command = 'docker cp {0}:{1} {2}'.format(
            docker_id,
            remote_filename,
            temp_filename,
        )

        status, _, stderr = ssh.run_shell_command(
            state, host,
            docker_command,
            print_output=print_output,
            print_input=print_input,
        )

        ssh_status = ssh.get_file(state, host, temp_filename, filename_or_io)
    finally:
        remote_remove(state, host, temp_filename, print_output=print_output,
                      print_input=print_input)

    if not ssh_status:
        raise IOError('failed to copy file over ssh')

    if not status:
        raise IOError('\n'.join(stderr))

    if print_output:
        click.echo('{0}file downloaded from container: {1}'.format(
            host.print_prefix, remote_filename,
        ), err=True)

    return status


EXECUTION_CONNECTOR = True



import os

from tempfile import mkstemp

import click
import six

from pyinfra import local, logger
from pyinfra.api import QuoteString, StringCommand
from pyinfra.api.exceptions import ConnectError, InventoryError, PyinfraError
from pyinfra.api.util import get_file_io, memoize
from pyinfra.progress import progress_spinner

from .local import run_shell_command as run_local_shell_command
from .util import get_sudo_password, make_unix_command


@memoize
def show_warning():
    logger.warning('The @chroot connector is in beta!')


def make_names_data(directory=None):
    if not directory:
        raise InventoryError('No directory provided!')

    show_warning()

    yield '@chroot/{0}'.format(directory), {
        'chroot_directory': '/{0}'.format(directory.lstrip('/')),
    }, ['@chroot']


def connect(state, host):
    chroot_directory = host.data.chroot_directory

    try:
        with progress_spinner({'chroot run'}):
            local.shell(
                'chroot {0} ls'.format(chroot_directory), splitlines=True,
            )
    except PyinfraError as e:
        raise ConnectError(e.args[0])

    host.host_data['chroot_directory'] = chroot_directory
    return True


def run_shell_command(
    state,
    host,
    command,
    get_pty=False,
    timeout=None,
    stdin=None,
    success_exit_codes=None,
    print_output=False,
    print_input=False,
    return_combined_output=False,
    use_sudo_password=False,
    **command_kwargs
):

    if use_sudo_password:
        command_kwargs['use_sudo_password'] = get_sudo_password(
            state,
            host,
            use_sudo_password,
            run_shell_command=run_shell_command,
            put_file=put_file,
        )

    chroot_directory = host.host_data['chroot_directory']

    command = make_unix_command(command, **command_kwargs)
    command = QuoteString(command)

    logger.debug(
        '--> Running chroot command on ({0}):{1}'.format(
            chroot_directory, command,
        ),
    )

    chroot_command = StringCommand(
        'chroot', chroot_directory,
        'sh', '-c', command,
    )

    return run_local_shell_command(
        state,
        host,
        chroot_command,
        timeout=timeout,
        stdin=stdin,
        success_exit_codes=success_exit_codes,
        print_output=print_output,
        print_input=print_input,
        return_combined_output=return_combined_output,
    )


def put_file(
    state,
    host,
    filename_or_io,
    remote_filename,
    print_output=False,
    print_input=False,
    **kwargs  # ignored (sudo/etc)
):

    _, temp_filename = mkstemp()

    try:
        # Load our file or IO object and write it to the temporary file
        with get_file_io(filename_or_io) as file_io:
            with open(temp_filename, 'wb') as temp_f:
                data = file_io.read()

                if isinstance(data, six.text_type):
                    data = data.encode()

                temp_f.write(data)

        chroot_directory = host.host_data['chroot_directory']

        chroot_command = 'cp {0} {1}/{2}'.format(
            temp_filename, chroot_directory, remote_filename,
        )

        status, _, stderr = run_local_shell_command(
            state,
            host,
            chroot_command,
            print_output=print_output,
            print_input=print_input,
        )
    finally:
        os.remove(temp_filename)

    if not status:
        raise IOError('\n'.join(stderr))

    if print_output:
        click.echo(
            '{0}file uploaded to chroot: {1}'.format(
                host.print_prefix, remote_filename,
            ),
            err=True,
        )

    return status


def get_file(
    state,
    host,
    remote_filename,
    filename_or_io,
    print_output=False,
    print_input=False,
    **kwargs  # ignored (sudo/etc)
):

    _, temp_filename = mkstemp()

    try:
        chroot_directory = host.host_data['chroot_directory']
        chroot_command = 'cp {0}/{1} {2}'.format(
            chroot_directory, remote_filename, temp_filename,
        )

        status, _, stderr = run_local_shell_command(
            state,
            host,
            chroot_command,
            print_output=print_output,
            print_input=print_input,
        )

        # Load the temporary file and write it to our file or IO object
        with open(temp_filename) as temp_f:
            with get_file_io(filename_or_io, 'wb') as file_io:
                data = temp_f.read()

                if isinstance(data, six.text_type):
                    data = data.encode()

                file_io.write(data)
    finally:
        os.remove(temp_filename)

    if not status:
        raise IOError('\n'.join(stderr))

    if print_output:
        click.echo(
            '{0}file downloaded from chroot: {1}'.format(
                host.print_prefix, remote_filename,
            ),
            err=True,
        )

    return status


EXECUTION_CONNECTOR = True



from mock import MagicMock, patch

from pyinfra.api import Config, State
from pyinfra.api.connect import connect_all
from pyinfra.api.exceptions import PyinfraError
from pyinfra.api.facts import get_facts

from ..paramiko_util import (
    PatchSSHTestCase,
)
from ..util import make_inventory


class TestOperationsApi(PatchSSHTestCase):
    def test_get_fact(self):
        inventory = make_inventory(hosts=('anotherhost',))
        state = State(inventory, Config())

        anotherhost = inventory.get_host('anotherhost')

        connect_all(state)

        with patch('pyinfra.api.connectors.ssh.run_shell_command') as fake_run_command:
            fake_run_command.return_value = MagicMock(), [('stdout', 'some-output')]
            fact_data = get_facts(state, 'command', ('yes',))

        assert fact_data == {anotherhost: 'some-output'}

        fake_run_command.assert_called_with(
            state,
            anotherhost,
            'yes',
            print_input=False,
            print_output=False,
            shell_executable=None,
            su_user=None,
            sudo=False,
            sudo_user=None,
            timeout=None,
            use_sudo_password=False,
            return_combined_output=True,
        )

    def test_get_fact_current_op_meta(self):
        inventory = make_inventory(hosts=('anotherhost',))
        state = State(inventory, Config())

        anotherhost = inventory.get_host('anotherhost')

        connect_all(state)
        state.current_op_global_kwargs = {
            'sudo': True,
            'sudo_user': 'someuser',
            'use_sudo_password': True,
            'su_user': 'someuser',
            'ignore_errors': False,
            'timeout': 10,
        }

        with patch('pyinfra.api.connectors.ssh.run_shell_command') as fake_run_command:
            fake_run_command.return_value = MagicMock(), [('stdout', 'some-output')]
            fact_data = get_facts(state, 'command', ('yes',))

        assert fact_data == {anotherhost: 'some-output'}

        fake_run_command.assert_called_with(
            state,
            anotherhost,
            'yes',
            print_input=False,
            print_output=False,
            shell_executable=None,
            su_user='someuser',
            sudo=True,
            sudo_user='someuser',
            timeout=10,
            use_sudo_password=True,
            return_combined_output=True,
        )

    def test_get_fact_error(self):
        inventory = make_inventory(hosts=('anotherhost',))
        state = State(inventory, Config())

        anotherhost = inventory.get_host('anotherhost')

        connect_all(state)

        with patch('pyinfra.api.connectors.ssh.run_shell_command') as fake_run_command:
            fake_run_command.return_value = False, MagicMock()

            with self.assertRaises(PyinfraError) as context:
                get_facts(state, 'command', ('fail command',))

        assert context.exception.args[0] == 'No hosts remaining!'

        fake_run_command.assert_called_with(
            state,
            anotherhost,
            'fail command',
            print_input=False,
            print_output=False,
            shell_executable=None,
            su_user=None,
            sudo=False,
            sudo_user=None,
            timeout=None,
            use_sudo_password=False,
            return_combined_output=True,
        )

    def test_get_fact_error_ignore(self):
        inventory = make_inventory(hosts=('anotherhost',))
        state = State(inventory, Config())

        anotherhost = inventory.get_host('anotherhost')

        connect_all(state)
        state.current_op_global_kwargs = {
            'sudo': False,
            'sudo_user': None,
            'use_sudo_password': False,
            'su_user': None,
            'ignore_errors': True,
            'timeout': None,
        }

        with patch('pyinfra.api.connectors.ssh.run_shell_command') as fake_run_command:
            fake_run_command.return_value = False, MagicMock()
            fact_data = get_facts(state, 'command', ('fail command',))

        assert fact_data == {anotherhost: None}

        fake_run_command.assert_called_with(
            state,
            anotherhost,
            'fail command',
            print_input=False,
            print_output=False,
            shell_executable=None,
            su_user=None,
            sudo=False,
            sudo_user=None,
            timeout=None,
            use_sudo_password=False,
            return_combined_output=True,
        )



# encoding: utf-8

from __future__ import unicode_literals

from unittest import TestCase

try:
    from pathlib import PurePosixPath
    HAS_PATHLIB = True
except ImportError:
    HAS_PATHLIB = False

from pyinfra.api.connectors.util import (
    escape_unix_path,
    make_unix_command,
    split_combined_output,
)


class TestConnectorUtil(TestCase):
    def test_split_combined_output_works(self):
        results = split_combined_output([
            ('stdout', 'stdout1'),
            ('stdout', 'stdout2'),
            ('stderr', 'stderr1'),
            ('stdout', 'stdout3'),
        ])

        assert results == (['stdout1', 'stdout2', 'stdout3'], ['stderr1'])

    def test_split_combined_output_raises(self):
        with self.assertRaises(ValueError):
            split_combined_output(['nope', ''])


class TestEscapeUnixPathUtil(TestCase):
    def test_path(self):
        escaped_path = escape_unix_path('/path/to/directory with space/ starts')
        assert escaped_path == '/path/to/directory\\ with\\ space/\\ starts'

    def test_escaped_path(self):
        escaped_path = '/path/to/directory\\ with\\ space/\\ starts'
        double_escaped_path = escape_unix_path(escaped_path)
        assert double_escaped_path == escaped_path

    def test_pathlib_path(self):
        if not HAS_PATHLIB:
            return

        path = PurePosixPath('/', 'path', 'to', 'directory with space', ' starts')
        escaped_path = escape_unix_path(path)
        assert escaped_path == '/path/to/directory\\ with\\ space/\\ starts'


class TestMakeUnixCommandConnectorUtil(TestCase):
    def test_command(self):
        command = make_unix_command('echo Šablony')
        assert command.get_raw_value() == "sh -c 'echo Šablony'"

    def test_sudo_command(self):
        command = make_unix_command('uptime', sudo=True)
        assert command.get_raw_value() == 'sudo -H -n sh -c uptime'

    def test_sudo_multi_arg_command(self):
        command = make_unix_command('echo hi', sudo=True, preserve_sudo_env=True)
        assert command.get_raw_value() == "sudo -H -n -E sh -c 'echo hi'"

    def test_sudo_preserve_env_command(self):
        command = make_unix_command('uptime', sudo=True, preserve_sudo_env=True)
        assert command.get_raw_value() == 'sudo -H -n -E sh -c uptime'

    def test_use_sudo_login_command(self):
        command = make_unix_command('uptime', sudo=True, use_sudo_login=True)
        assert command.get_raw_value() == 'sudo -H -n -i sh -c uptime'

    def test_sudo_user_command(self):
        command = make_unix_command('uptime', sudo=True, sudo_user='pyinfra')
        assert command.get_raw_value() == 'sudo -H -n -u pyinfra sh -c uptime'

    def test_su_command(self):
        command = make_unix_command('uptime', su_user='pyinfra')
        assert command.get_raw_value() == "su pyinfra -c 'sh -c uptime'"

    def test_su_multi_arg_command(self):
        command = make_unix_command('echo hi', su_user='pyinfra')
        assert command.get_raw_value() == "su pyinfra -c 'sh -c '\"'\"'echo hi'\"'\"''"

    def test_use_su_login_command(self):
        command = make_unix_command('uptime', su_user='pyinfra', use_su_login=True)
        assert command.get_raw_value() == "su -l pyinfra -c 'sh -c uptime'"

    def test_preserve_su_env_command(self):
        command = make_unix_command('uptime', su_user='pyinfra', preserve_su_env=True)
        assert command.get_raw_value() == "su -m pyinfra -c 'sh -c uptime'"

    def test_su_shell_command(self):
        command = make_unix_command('uptime', su_user='pyinfra', su_shell='bash')
        assert command.get_raw_value() == "su -s `which bash` pyinfra -c 'sh -c uptime'"

    def test_command_env(self):
        command = make_unix_command('uptime', env={
            'key': 'value',
            'anotherkey': 'anothervalue',
        })
        assert command.get_raw_value() in [
            "sh -c 'env key=value anotherkey=anothervalue uptime'",
            "sh -c 'env anotherkey=anothervalue key=value uptime'",
        ]

    def test_command_chdir(self):
        command = make_unix_command('uptime', chdir='/opt/somedir')
        assert command.get_raw_value() == "sh -c 'cd /opt/somedir && uptime'"

    def test_custom_shell_command(self):
        command = make_unix_command('uptime', shell_executable='bash')
        assert command.get_raw_value() == 'bash -c uptime'

    def test_mixed_command(self):
        command = make_unix_command(
            'echo hi',
            chdir='/opt/somedir',
            env={'key': 'value'},
            sudo=True,
            sudo_user='root',
            preserve_sudo_env=True,
            su_user='pyinfra',
            shell_executable='bash',
        )
        assert command.get_raw_value() == (
            'sudo -H -n -E -u root '  # sudo bit
            'su pyinfra -c '  # su bit
            "'bash -c '\"'\"'cd /opt/somedir && env key=value echo hi'\"'\"''"  # command bit
        )



from __future__ import unicode_literals

from datetime import datetime


def ensure_mode_int(mode):
    # Already an int (/None)?
    if isinstance(mode, int) or mode is None:
        return mode

    try:
        # Try making an int ('700' -> 700)
        return int(mode)

    except (TypeError, ValueError):
        pass

    # Return as-is (ie +x which we don't need to normalise, it always gets run)
    return mode


def ensure_whole_line_match(line):
    # Ensure we're matching a whole ^line$
    if not line.startswith('^'):
        line = '^.*{0}'.format(line)

    if not line.endswith('$'):
        line = '{0}.*$'.format(line)

    return line


def get_timestamp():
    return datetime.now().strftime('%y%m%d%H%M')


def sed_replace(
    filename,
    line,
    replace,
    flags=None,
    backup=False,
    interpolate_variables=False,
):
    flags = ''.join(flags) if flags else ''

    line = line.replace('/', r'\/')
    replace = str(replace)
    replace = replace.replace('/', r'\/')
    backup_extension = get_timestamp()

    string_to_format = (
        'sed -i.{backup_extension} "s/{0}/{1}/{2}" {3}'
        if interpolate_variables else
        "sed -i.{backup_extension} 's/{0}/{1}/{2}' {3}"
    )

    sed_command = string_to_format.format(
        line, replace, flags, filename,
        backup_extension=backup_extension,
    )

    if not backup:  # if we're not backing up, remove the file *if* sed succeeds
        sed_command = '{0} && rm -f {1}.{2}'.format(sed_command, filename, backup_extension)

    return sed_command


def chmod(target, mode, recursive=False):
    return 'chmod {0}{1} {2}'.format(('-R ' if recursive else ''), mode, target)


def chown(target, user, group=None, recursive=False, dereference=True):
    command = 'chown'
    user_group = None

    if user and group:
        user_group = '{0}:{1}'.format(user, group)

    elif user:
        user_group = user

    elif group:
        command = 'chgrp'
        user_group = group

    return '{0}{1}{2} {3} {4}'.format(
        command,
        ' -R' if recursive else '',
        ' -h' if not dereference else '',
        user_group,
        target,
    )



from six.moves.urllib.parse import urlparse

from pyinfra.api import FactBase


class GpgFactBase(FactBase):
    abstract = True

    requires_command = 'gpg'

    key_record_type = 'pub'
    subkey_record_type = 'sub'

    def process(self, output):
        # For details on the field values see:
        # https://git.gnupg.org/cgi-bin/gitweb.cgi?p=gnupg.git;a=blob_plain;f=doc/DETAILS
        keys = []

        current_key = None
        current_subkey = None

        for line in output:
            if not line:
                continue

            bits = line.split(':')

            if bits[0] in (self.key_record_type, self.subkey_record_type):
                key_details = {
                    'validity': bits[1],
                    'length': int(bits[2]),
                    'id': bits[4],
                }

                if bits[0] == self.key_record_type:
                    key_details['subkeys'] = []

                    if current_key:
                        if current_subkey:
                            current_key['subkeys'].append(current_subkey)

                        keys.append(current_key)

                    current_key = key_details
                    current_subkey = None
                elif current_key:
                    current_subkey = key_details

            elif current_subkey or current_key:
                target = current_subkey or current_key

                if bits[0] == 'fpr':
                    target['fingerprint'] = bits[9]  # fingerprint = field 10
                elif bits[0] == 'uid':
                    target['uid_hash'] = bits[7]
                    target['uid'] = bits[9]

        if current_key:
            if current_subkey:
                current_key['subkeys'].append(current_subkey)

            keys.append(current_key)

        # Return as a dictionary of keyID -> details
        keys_by_id = {}
        for key in keys:
            if 'subkeys' in key:
                key['subkeys'] = {
                    subkey.pop('id'): subkey
                    for subkey in key['subkeys']
                }
            keys_by_id[key.pop('id')] = key

        return keys_by_id


class GpgKey(GpgFactBase):
    '''
    Returns information on one or more GPG keys found in a file or URL.

    .. code:: python

        {
            'KEY-ID': {
                'length': 4096,
                'uid': 'Oxygem <hello@oxygem.com>'
            },
            ...
        }
    '''

    def command(self, src):
        if urlparse(src).scheme:
            return (
                '(wget -O - {0} || curl -sSLf {0}) | gpg --with-colons'
            ).format(src)

        return 'gpg --with-colons {0}'.format(src)


class GpgKeys(GpgFactBase):
    '''
    Returns information on all public keys in a keychain.

    .. code:: python

        {
            'KEY-ID': {
                'length': 4096,
                'uid': 'Oxygem <hello@oxygem.com>'
            },
            ...
        }
    '''

    def command(self, keyring=None):
        if not keyring:
            return 'gpg --list-keys --with-colons'

        return (
            'gpg --list-keys --with-colons --keyring {0} --no-default-keyring'
        ).format(keyring)


class GpgSecretKeys(GpgFactBase):
    '''
    Returns information on all secret keys in a keychain.

    .. code:: python

        {
            'KEY-ID': {
                'length': 4096,
                'fingerprint': 'ABC',
                'uid': 'Oxygem <hello@oxygem.com>'
            },
            ...
        }
    '''

    key_record_type = 'sec'
    subkey_record_type = 'ssb'

    def command(self, keyring=None):
        if not keyring:
            return 'gpg --list-secret-keys --with-colons'

        return (
            'gpg --list-secret-keys --with-colons --keyring {0} --no-default-keyring'
        ).format(keyring)
